# pylint: disable=too-many-lines
import os
import warnings
from typing import Union, Callable, List, Dict, Tuple, Any

try:
    import pandas as pd
except ImportError:
    pd = None

try:
    # pylint: disable=unused-import
    import matplotlib.pyplot

    matplotlib = True
except ImportError:
    matplotlib = False

import torch

try:
    from torch.utils.tensorboard import SummaryWriter
except ImportError:
    SummaryWriter = None

from . import Model
from ..utils import set_seeds
from ..plotting import plot_history
from .callbacks import (
    ModelCheckpoint,
    OptimizerCheckpoint,
    LRSchedulerCheckpoint,
    PeriodicSaveLambda,
    AtomicCSVLogger,
    TensorBoardLogger,
    BestModelRestore,
)


class Experiment:
    """
    The Experiment class provides a straightforward experimentation tool for efficient and entirely
    customizable finetuning of the whole neural network training procedure with PyTorch. The
    ``Experiment`` object takes care of the training and testing processes while also managing to
    keep traces of all pertinent information via the automatic logging option.

    Args:
        directory (str): Path to the experiment's working directory. Will be used for automatic logging.
        network (torch.nn.Module): A PyTorch network.
        device (Union[torch.torch.device, List[torch.torch.device], str, None]): The device to which the model is sent
            or for multi-GPUs, the list of devices to which the model is to be sent. When using a string for a multiple
            GPUs, the option is "all", for "take them all." By default, the current device is used as the main one.
            If None, the model will be kept on its current device.
            (Default value = None)
        logging (bool): Whether or not to log the experiment's progress. If true, various logging
            callbacks will be inserted to output training and testing stats as well as to save model checkpoints,
            for example, automatically. See :func:`~Experiment.train()` and :func:`~Experiment.test()` for more details.
            (Default value = True)
        optimizer (Union[torch.optim.Optimizer, str]): If Pytorch Optimizer, must already be initialized.
            If str, should be the optimizer's name in Pytorch (i.e. 'Adam' for torch.optim.Adam).
            (Default value = 'sgd')
        loss_function(Union[Callable, str], optional) It can be any PyTorch
            loss layer or custom loss function. It can also be a string with the same name as a PyTorch
            loss function (either the functional or object name). The loss function must have the signature
            ``loss_function(input, target)`` where ``input`` is the prediction of the network and ``target``
            is the ground truth. If ``None``, will default to, in priority order, either the model's own
            loss function or the default loss function associated with the ``task``.
            (Default value = None)
        batch_metrics (List, optional): List of functions with the same signature as the loss function. Each metric
            can be any PyTorch loss function. It can also be a string with the same name as a PyTorch
            loss function (either the functional or object name). 'accuracy' (or just 'acc') is also a
            valid metric. Each metric function is called on each batch of the optimization and on the
            validation batches at the end of the epoch.
            (Default value = None)
        epoch_metrics (List, optional): List of functions with the same signature as
            :class:`~poutyne.EpochMetric`
            (Default value = None)
        monitoring (bool): Whether or not to monitor the training. If True will track the best epoch.
            If False, ``monitor_metric`` and ``monitor_mode`` are not used, and when testing, the last epoch is used to
            test the model instead of the best epoch.
            (Default value = True)
        monitor_metric (str, optional): Which metric to consider for best model performance calculation. Should be in
            the format '{metric_name}' or 'val_{metric_name}' (i.e. 'val_loss'). If None, will follow the value
            suggested by ``task`` or default to 'val_loss'. If ``monitoring`` is set to False, will be ignore.

            .. warning:: If you do not plan on using a validation set, you must set the monitor metric to another
                value.
        monitor_mode (str, optional): Which mode, either 'min' or 'max', should be used when considering the
            ``monitor_metric`` value. If None, will follow the value suggested by ``task`` or default to 'min'.
            If ``monitoring`` is set to False, will be ignore.
        task (str, optional): Any str beginning with either 'classif' or 'reg'. Specifying a ``task``
            can assign default values to the ``loss_function``, ``batch_metrics``, ``monitor_mode`` and
            ``monitor_mode``. For ``task`` that begins with 'reg', the only default value is the loss function
            that is the mean squared error. When beginning with 'classif', the default loss function is the
            cross-entropy loss. The default batch metrics will be the accuracy, the default epoch metrics will be
            the F1 score and the default monitoring will be set on 'val_acc' with a 'max' mode.
            (Default value = None)

    Examples:
        Using a PyTorch DataLoader, on classification task with SGD optimizer::

            import torch
            from torch.utils.data import DataLoader, TensorDataset
            from poutyne import Experiment

            num_features = 20
            num_classes = 5

            # Our training dataset with 800 samples.
            num_train_samples = 800
            train_x = torch.rand(num_train_samples, num_features)
            train_y = torch.randint(num_classes, (num_train_samples, ), dtype=torch.long)
            train_dataset = TensorDataset(train_x, train_y)
            train_generator = DataLoader(train_dataset, batch_size=32)

            # Our validation dataset with 200 samples.
            num_valid_samples = 200
            valid_x = torch.rand(num_valid_samples, num_features)
            valid_y = torch.randint(num_classes, (num_valid_samples, ), dtype=torch.long)
            valid_dataset = TensorDataset(valid_x, valid_y)
            valid_generator = DataLoader(valid_dataset, batch_size=32)

            # Our network
            pytorch_network = torch.nn.Linear(num_features, num_train_samples)

            # Initialization of our experimentation and network training
            exp = Experiment('./simple_example',
                             pytorch_network,
                             optimizer='sgd',
                             task='classif')
            exp.train(train_generator, valid_generator, epochs=5)

    The above code will yield an output similar to the below lines. Note the automatic checkpoint saving
    in the experiment directory when the monitored metric improved.

    .. code-block:: none

            Epoch 1/5 0.09s Step 25/25: loss: 6.351375, acc: 1.375000, val_loss: 6.236106, val_acc: 5.000000
            Epoch 1: val_acc improved from -inf to 5.00000, saving file to ./simple_example/checkpoint_epoch_1.ckpt
            Epoch 2/5 0.10s Step 25/25: loss: 6.054254, acc: 14.000000, val_loss: 5.944495, val_acc: 19.500000
            Epoch 2: val_acc improved from 5.00000 to 19.50000, saving file to ./simple_example/checkpoint_epoch_2.ckpt
            Epoch 3/5 0.09s Step 25/25: loss: 5.759377, acc: 22.875000, val_loss: 5.655412, val_acc: 21.000000
            Epoch 3: val_acc improved from 19.50000 to 21.00000, saving file to ./simple_example/checkpoint_epoch_3.ckpt
            ...

    Training can now easily be resumed from the best checkpoint::

            exp.train(train_generator, valid_generator, epochs=10)

    .. code-block:: none

            Restoring model from ./simple_example/checkpoint_epoch_3.ckpt
            Loading weights from ./simple_example/checkpoint.ckpt and starting at epoch 6.
            Loading optimizer state from ./simple_example/checkpoint.optim and starting at epoch 6.
            Epoch 6/10 0.16s Step 25/25: loss: 4.897135, acc: 22.875000, val_loss: 4.813141, val_acc: 20.500000
            Epoch 7/10 0.10s Step 25/25: loss: 4.621514, acc: 22.625000, val_loss: 4.545359, val_acc: 20.500000
            Epoch 8/10 0.24s Step 25/25: loss: 4.354721, acc: 23.625000, val_loss: 4.287117, val_acc: 20.500000
            ...

    Testing is also very intuitive::

            exp.test(test_generator)

    .. code-block:: none

            Restoring model from ./simple_example/checkpoint_epoch_9.ckpt
            Found best checkpoint at epoch: 9
            lr: 0.01, loss: 4.09892, acc: 23.625, val_loss: 4.04057, val_acc: 21.5
            On best model: test_loss: 4.06664, test_acc: 17.5


    Finally, all the pertinent metrics specified to the Experiment at each epoch are stored in a specific logging
    file, found here at './simple_example/log.tsv'.

    .. code-block:: none

            epoch	time	            lr	    loss	            acc	    val_loss	        val_acc
            1	    0.0721172170015052	0.01	6.351375141143799	1.375	6.23610631942749	5.0
            2	    0.0298177790245972	0.01	6.054253826141357	14.000	5.94449516296386	19.5
            3	    0.0637106419890187	0.01	5.759376544952392	22.875	5.65541223526001	21.0
            ...

    Also, we could use more than one GPU (on a single node) by using the device argument

    .. code-block:: none

            # Initialization of our experimentation and network training
            exp = Experiment('./simple_example',
                             pytorch_network,
                             optimizer='sgd',
                             task='classif',
                             device="all")
            exp.train(train_generator, valid_generator, epochs=5)

    """

    BEST_CHECKPOINT_FILENAME = 'checkpoint_epoch_{epoch}.ckpt'
    MODEL_CHECKPOINT_FILENAME = 'checkpoint.ckpt'
    OPTIMIZER_CHECKPOINT_FILENAME = 'checkpoint.optim'
    LOG_FILENAME = 'log.tsv'
    TENSORBOARD_DIRECTORY = 'tensorboard'
    EPOCH_FILENAME = 'last.epoch'
    LR_SCHEDULER_FILENAME = 'lr_sched_%d.lrsched'
    PLOTS_DIRECTORY = 'plots'
    TEST_LOG_FILENAME = '{name}_log.tsv'

    def __init__(
        self,
        directory: str,
        network: torch.nn.Module,
        *,
        device: Union[torch.device, List[torch.device], List[str], None, str] = None,
        logging: bool = True,
        optimizer: Union[torch.optim.Optimizer, str] = 'sgd',
        loss_function: Union[Callable, str] = None,
        batch_metrics: Union[List, None] = None,
        epoch_metrics: Union[List, None] = None,
        monitoring: bool = True,
        monitor_metric: Union[str, None] = None,
        monitor_mode: Union[str, None] = None,
        task: Union[str, None] = None,
    ) -> None:
        if pd is None:
            raise ImportError("pandas needs to be installed to use the class Experiment.")

        self.directory = directory
        self.logging = logging

        if task is not None and not task.startswith('classif') and not task.startswith('reg'):
            raise ValueError(f"Invalid task '{task}'")

        batch_metrics = [] if batch_metrics is None else batch_metrics
        epoch_metrics = [] if epoch_metrics is None else epoch_metrics

        loss_function = self._get_loss_function(loss_function, network, task)
        batch_metrics = self._get_batch_metrics(batch_metrics, network, task)
        epoch_metrics = self._get_epoch_metrics(epoch_metrics, network, task)

        self.monitoring = monitoring
        self.monitor_metric = None
        self.monitor_mode = None
        if self.monitoring:
            self._set_monitor(monitor_metric, monitor_mode, task)

        self.model = Model(
            network, optimizer, loss_function, batch_metrics=batch_metrics, epoch_metrics=epoch_metrics, device=device
        )

        self.best_checkpoint_filename = self.get_path(Experiment.BEST_CHECKPOINT_FILENAME)
        self.model_checkpoint_filename = self.get_path(Experiment.MODEL_CHECKPOINT_FILENAME)
        self.optimizer_checkpoint_filename = self.get_path(Experiment.OPTIMIZER_CHECKPOINT_FILENAME)
        self.log_filename = self.get_path(Experiment.LOG_FILENAME)
        self.tensorboard_directory = self.get_path(Experiment.TENSORBOARD_DIRECTORY)
        self.epoch_filename = self.get_path(Experiment.EPOCH_FILENAME)
        self.lr_scheduler_filename = self.get_path(Experiment.LR_SCHEDULER_FILENAME)
        self.plots_directory = self.get_path(Experiment.PLOTS_DIRECTORY)
        self.test_log_filename = self.get_path(Experiment.TEST_LOG_FILENAME)

    def get_path(self, *paths: str) -> str:
        """
        Returns the path inside the experiment directory.
        """
        return os.path.join(self.directory, *paths)

    def _get_loss_function(
        self, loss_function: Union[Callable, str], network: torch.nn.Module, task: Union[str, None]
    ) -> Union[Callable, str]:
        if loss_function is None:
            if hasattr(network, 'loss_function'):
                return network.loss_function
            if task is not None:
                if task.startswith('classif'):
                    return 'cross_entropy'
                if task.startswith('reg'):
                    return 'mse'
        return loss_function

    def _get_batch_metrics(
        self, batch_metrics: Union[List, None], network: torch.nn.Module, task: Union[str, None]
    ) -> Union[List, None]:
        if batch_metrics is None or len(batch_metrics) == 0:
            if hasattr(network, 'batch_metrics'):
                return network.batch_metrics
            if task is not None and task.startswith('classif'):
                return ['accuracy']
        return batch_metrics

    def _get_epoch_metrics(
        self, epoch_metrics: Union[List, None], network, task: Union[str, None]
    ) -> Union[List, None]:
        if epoch_metrics is None or len(epoch_metrics) == 0:
            if hasattr(network, 'epoch_metrics'):
                return network.epoch_metrics
            if task is not None and task.startswith('classif'):
                return ['f1']
        return epoch_metrics

    def _set_monitor(
        self, monitor_metric: Union[str, None], monitor_mode: Union[str, None], task: Union[str, None]
    ) -> None:
        if monitor_mode is not None and monitor_mode not in ['min', 'max']:
            raise ValueError(f"Invalid mode '{monitor_mode}'")

        self.monitor_metric = 'val_loss'
        self.monitor_mode = 'min'
        if monitor_metric is not None:
            self.monitor_metric = monitor_metric
            if monitor_mode is not None:
                self.monitor_mode = monitor_mode
        elif task is not None and task.startswith('classif'):
            self.monitor_metric = 'val_acc'
            self.monitor_mode = 'max'

    def get_stats(self):
        return pd.read_csv(self.log_filename, sep='\t')

    def get_best_epoch_stats(self) -> Dict:
        """
        Returns all computed statistics corresponding to the best epoch according to the
        ``monitor_metric`` and ``monitor_mode`` attributes.

        Returns:
            dict where each key is a column name in the logging output file
            and values are the ones found at the best epoch.
        """
        if not self.monitoring:
            raise ValueError("Monitoring was disabled. Cannot get best epoch.")

        history = self.get_stats()
        if self.monitor_mode == 'min':
            best_epoch_index = history[self.monitor_metric].idxmin()
        else:
            best_epoch_index = history[self.monitor_metric].idxmax()
        return history.iloc[best_epoch_index : best_epoch_index + 1]

    def get_saved_epochs(self):
        """
        Returns a pandas DataFrame which each row corresponds to an epoch having
        a saved checkpoint.

        Returns:
            pandas DataFrame which each row corresponds to an epoch having a saved
            checkpoint.
        """
        if not self.monitoring:
            raise ValueError("Monitoring was disabled. Except the last epoch, no epoch checkpoint were saved.")

        history = self.get_stats()
        metrics = history[self.monitor_metric].tolist()
        if self.monitor_mode == 'min':
            monitor_op = lambda x, y: x < y
            current_best = float('Inf')
        elif self.monitor_mode == 'max':
            monitor_op = lambda x, y: x > y
            current_best = -float('Inf')
        saved_epoch_indices = []
        for i, metric in enumerate(metrics):
            if monitor_op(metric, current_best):
                current_best = metric
                saved_epoch_indices.append(i)
        return history.iloc[saved_epoch_indices]

    def _warn_missing_file(self, filename: str) -> None:
        warnings.warn(f"Missing checkpoint: {filename}.")

    def _load_epoch_state(self, lr_schedulers: List) -> int:
        # pylint: disable=broad-except
        initial_epoch = 1
        if os.path.isfile(self.epoch_filename):
            try:
                with open(self.epoch_filename, 'r', encoding='utf-8') as f:
                    initial_epoch = int(f.read()) + 1
            except Exception as e:
                print(e)
            if os.path.isfile(self.model_checkpoint_filename):
                try:
                    print(
                        f"Loading weights from {self.model_checkpoint_filename} and "
                        f"starting at epoch {initial_epoch:d}."
                    )
                    self.model.load_weights(self.model_checkpoint_filename)
                except Exception as e:
                    print(e)
            else:
                self._warn_missing_file(self.model_checkpoint_filename)
            if os.path.isfile(self.optimizer_checkpoint_filename):
                try:
                    print(
                        f"Loading optimizer state from {self.optimizer_checkpoint_filename} and "
                        f"starting at epoch {initial_epoch:d}."
                    )
                    self.model.load_optimizer_state(self.optimizer_checkpoint_filename)
                except Exception as e:
                    print(e)
            else:
                self._warn_missing_file(self.optimizer_checkpoint_filename)
            for i, lr_scheduler in enumerate(lr_schedulers):
                filename = self.lr_scheduler_filename % i
                if os.path.isfile(filename):
                    try:
                        print(f"Loading LR scheduler state from {filename} and starting at epoch {initial_epoch:d}.")
                        lr_scheduler.load_state(filename)
                    except Exception as e:
                        print(e)
                else:
                    self._warn_missing_file(filename)
        return initial_epoch

    def _init_model_restoring_callbacks(
        self, initial_epoch: int, keep_only_last_best: bool, save_every_epoch: bool
    ) -> List:
        callbacks = []
        best_checkpoint = ModelCheckpoint(
            self.best_checkpoint_filename,
            monitor=self.monitor_metric,
            mode=self.monitor_mode,
            keep_only_last_best=keep_only_last_best,
            save_best_only=not save_every_epoch,
            restore_best=not save_every_epoch,
            verbose=not save_every_epoch,
        )
        callbacks.append(best_checkpoint)

        if save_every_epoch:
            best_restore = BestModelRestore(monitor=self.monitor_metric, mode=self.monitor_mode, verbose=True)
            callbacks.append(best_restore)

        if initial_epoch > 1:
            # We set the current best metric score in the ModelCheckpoint so that
            # it does not save checkpoint it would not have saved if the
            # optimization was not stopped.
            best_epoch_stats = self.get_best_epoch_stats()
            best_epoch = best_epoch_stats['epoch'].item()
            best_filename = self.best_checkpoint_filename.format(epoch=best_epoch)
            if not save_every_epoch:
                best_checkpoint.best_filename = best_filename
                best_checkpoint.current_best = best_epoch_stats[self.monitor_metric].item()
            else:
                best_restore.best_weights = torch.load(best_filename, map_location='cpu')
                best_restore.current_best = best_epoch_stats[self.monitor_metric].item()

        return callbacks

    def _init_tensorboard_callbacks(self, disable_tensorboard: bool) -> Tuple:
        tensorboard_writer = None
        callbacks = []
        if not disable_tensorboard:
            if SummaryWriter is None:
                warnings.warn(
                    "tensorboard does not seem to be installed. "
                    "To remove this warning, set the 'disable_tensorboard' "
                    "flag to True or install tensorboard.",
                    stacklevel=3,
                )
            else:
                tensorboard_writer = SummaryWriter(self.tensorboard_directory)
                callbacks += [TensorBoardLogger(tensorboard_writer)]
        return tensorboard_writer, callbacks

    def _init_lr_scheduler_callbacks(self, lr_schedulers: List) -> List:
        callbacks = []
        if self.logging:
            for i, lr_scheduler in enumerate(lr_schedulers):
                filename = self.lr_scheduler_filename % i
                callbacks += [LRSchedulerCheckpoint(lr_scheduler, filename, verbose=False)]
        else:
            callbacks += lr_schedulers
        return callbacks

    def _save_history(self):
        if matplotlib:
            history = self.get_stats()
            plot_history(
                history,
                show=False,
                save=True,
                close=True,
                save_directory=self.plots_directory,
                save_extensions=('png', 'pdf'),
            )

    def train(self, train_generator, valid_generator=None, **kwargs) -> List[Dict]:
        """
        Trains or finetunes the model on a dataset using a generator. If a previous training already occurred
        and lasted a total of `n_previous` epochs, then the model's weights will be set to the last checkpoint and the
        training will be resumed for epochs range (`n_previous`, `epochs`].

        If the Experiment has logging enabled (i.e. self.logging is True), numerous callbacks will be automatically
        included. Notably, two :class:`~poutyne.ModelCheckpoint` objects will take care of saving the last and every
        new best (according to monitor mode) model weights in appropriate checkpoint files.
        :class:`~poutyne.OptimizerCheckpoint` and :class:`~poutyne.LRSchedulerCheckpoint` will also respectively
        handle the saving of the optimizer and LR scheduler's respective states for future retrieval. Moreover, a
        :class:`~poutyne.AtomicCSVLogger` will save all available epoch statistics in an output .tsv file. Lastly, a
        :class:`~poutyne.TensorBoardLogger` handles automatic TensorBoard logging of various neural network
        statistics.

        .. warning:: With **Jupyter Notebooks in Firefox**, if ``colorama`` is installed and colors are enabled (as it
            is by default), a great number of epochs and steps per epoch can cause a spike in memory usage in Firefox.
            The problem does not occur in Google Chrome/Chromium. To avoid this problem, you can disable the colors by
            passing ``progress_options={'coloring': False}``. See
            `this Github issue for details <https://github.com/jupyter/notebook/issues/5897>`__.

        Args:
            train_generator: Generator-like object for the training set. See :func:`~Model.fit_generator()`
                for details on the types of generators supported.
            valid_generator (optional): Generator-like object for the validation set. See
                :func:`~Model.fit_generator()` for details on the types of generators supported.
                (Default value = None)
            callbacks (List[~poutyne.Callback]): List of callbacks that will be called during
                training. These callbacks are added after those used in this method (see above). This allows to assume
                that they are called after those.
                (Default value = None)
            lr_schedulers: List of learning rate schedulers. (Default value = None)
            keep_only_last_best (bool): Whether only the last saved best checkpoint is kept. Applies only when
                 `save_every_epoch` is false.
                 (Default value = False)
            save_every_epoch (bool, optional): Whether or not to save the experiment model's weights after
                every epoch.
                (Default value = False)
            disable_tensorboard (bool, optional): Whether or not to disable the automatic tensorboard logging
                callbacks.
                (Default value = False)
            seed (int, optional): Seed used to make the sampling deterministic.
                (Default value = 42)
            kwargs: Any keyword arguments to pass to :func:`~Model.fit_generator()`.

        Returns:
            List of dict containing the history of each epoch.
        """
        return self._train(self.model.fit_generator, train_generator, valid_generator, **kwargs)

    def train_dataset(self, train_dataset, valid_dataset=None, **kwargs) -> List[Dict]:
        """
        Trains or finetunes the model on a dataset. If a previous training already occurred
        and lasted a total of `n_previous` epochs, then the model's weights will be set to the last checkpoint and the
        training will be resumed for epochs range (`n_previous`, `epochs`].

        If the Experiment has logging enabled (i.e. self.logging is True), numerous callbacks will be automatically
        included. Notably, two :class:`~poutyne.ModelCheckpoint` objects will take care of saving the last and every
        new best (according to monitor mode) model weights in appropriate checkpoint files.
        :class:`~poutyne.OptimizerCheckpoint` and :class:`~poutyne.LRSchedulerCheckpoint` will also respectively
        handle the saving of the optimizer and LR scheduler's respective states for future retrieval. Moreover, a
        :class:`~poutyne.AtomicCSVLogger` will save all available epoch statistics in an output .tsv file. Lastly, a
        :class:`~poutyne.TensorBoardLogger` handles automatic TensorBoard logging of various neural network
        statistics.

        Args:
            train_dataset (~torch.utils.data.Dataset): Training dataset.
            valid_dataset (~torch.utils.data.Dataset): Validation dataset.
            callbacks (List[~poutyne.Callback]): List of callbacks that will be called during
                training. These callbacks are added after those used in this method (see above). This allows to assume
                that they are called after those.
                (Default value = None)
            lr_schedulers: List of learning rate schedulers. (Default value = None)
            keep_only_last_best (bool): Whether only the last saved best checkpoint is kept. Applies only when
                 `save_every_epoch` is false.
                 (Default value = False)
            save_every_epoch (bool, optional): Whether or not to save the experiment model's weights after
                every epoch.
                (Default value = False)
            disable_tensorboard (bool, optional): Whether or not to disable the automatic tensorboard logging
                callbacks.
                (Default value = False)
            seed (int, optional): Seed used to make the sampling deterministic.
                (Default value = 42)
            kwargs: Any keyword arguments to pass to :func:`~Model.fit_dataset()`.

        Returns:
            List of dict containing the history of each epoch.
        """
        return self._train(self.model.fit_dataset, train_dataset, valid_dataset, **kwargs)

    def train_data(self, x, y, validation_data=None, **kwargs) -> List[Dict]:
        """
        Trains or finetunes the model on data under the form of NumPy arrays or torch tensors. If a previous
        training already occurred and lasted a total of `n_previous` epochs, then the model's weights will be set to the
        last checkpoint and the training will be resumed for epochs range (`n_previous`, `epochs`].

        If the Experiment has logging enabled (i.e. self.logging is True), numerous callbacks will be automatically
        included. Notably, two :class:`~poutyne.ModelCheckpoint` objects will take care of saving the last and every
        new best (according to monitor mode) model weights in appropriate checkpoint files.
        :class:`~poutyne.OptimizerCheckpoint` and :class:`~poutyne.LRSchedulerCheckpoint` will also respectively
        handle the saving of the optimizer and LR scheduler's respective states for future retrieval. Moreover, a
        :class:`~poutyne.AtomicCSVLogger` will save all available epoch statistics in an output .tsv file. Lastly, a
        :class:`~poutyne.TensorBoardLogger` handles automatic TensorBoard logging of various neural network
        statistics.

        Args:
            x (Union[~torch.Tensor, ~numpy.ndarray] or Union[tuple, list] of Union[~torch.Tensor, ~numpy.ndarray]):
                Training dataset. Union[Tensor, ndarray] if the model has a single input.
                Union[tuple, list] of Union[Tensor, ndarray] if the model has multiple inputs.
            y (Union[~torch.Tensor, ~numpy.ndarray] or Union[tuple, list] of Union[~torch.Tensor, ~numpy.ndarray]):
                Target. Union[Tensor, ndarray] if the model has a single output.
                Union[tuple, list] of Union[Tensor, ndarray] if the model has multiple outputs.
            validation_data (Tuple[``x_val``, ``y_val``]):
                Same format as ``x`` and ``y`` previously described. Validation dataset on which to
                evaluate the loss and any model metrics at the end of each epoch. The model will not be
                trained on this data.
                (Default value = None)
            callbacks (List[~poutyne.Callback]): List of callbacks that will be called during
                training. These callbacks are added after those used in this method (see above). This allows to assume
                that they are called after those.
                (Default value = None)
            lr_schedulers: List of learning rate schedulers. (Default value = None)
            keep_only_last_best (bool): Whether only the last saved best checkpoint is kept. Applies only when
                 `save_every_epoch` is false.
                 (Default value = False)
            save_every_epoch (bool, optional): Whether or not to save the experiment model's weights after
                every epoch.
                (Default value = False)
            disable_tensorboard (bool, optional): Whether or not to disable the automatic tensorboard logging
                callbacks.
                (Default value = False)
            seed (int, optional): Seed used to make the sampling deterministic.
                (Default value = 42)
            kwargs: Any keyword arguments to pass to :func:`~Model.fit()`.

        Returns:
            List of dict containing the history of each epoch.
        """
        return self._train(self.model.fit, x, y, validation_data, **kwargs)

    def _train(
        self,
        training_func,
        *args,
        callbacks: Union[List, None] = None,
        lr_schedulers: Union[List, None] = None,
        keep_only_last_best: bool = False,
        save_every_epoch: bool = False,
        disable_tensorboard: bool = False,
        seed: int = 42,
        **kwargs,
    ) -> List[Dict]:
        set_seeds(seed)

        lr_schedulers = [] if lr_schedulers is None else lr_schedulers

        expt_callbacks = []

        tensorboard_writer = None
        initial_epoch = 1
        if self.logging:
            if not os.path.exists(self.directory):
                os.makedirs(self.directory)

            # Restarting optimization if needed.
            initial_epoch = self._load_epoch_state(lr_schedulers)

            expt_callbacks += [AtomicCSVLogger(self.log_filename, separator='\t', append=initial_epoch != 1)]

            if self.monitoring:
                expt_callbacks += self._init_model_restoring_callbacks(
                    initial_epoch, keep_only_last_best, save_every_epoch
                )
            expt_callbacks += [ModelCheckpoint(self.model_checkpoint_filename, verbose=False)]
            expt_callbacks += [OptimizerCheckpoint(self.optimizer_checkpoint_filename, verbose=False)]

            # We save the last epoch number after the end of the epoch so that the
            # _load_epoch_state() knows which epoch to restart the optimization.
            expt_callbacks += [
                PeriodicSaveLambda(lambda fd, epoch, logs: print(epoch, file=fd), self.epoch_filename, open_mode='w')
            ]

            tensorboard_writer, cb_list = self._init_tensorboard_callbacks(disable_tensorboard)
            expt_callbacks += cb_list
        else:
            if self.monitoring:
                expt_callbacks += [BestModelRestore(monitor=self.monitor_metric, mode=self.monitor_mode, verbose=True)]

        # This method returns callbacks that checkpoints the LR scheduler if logging is enabled.
        # Otherwise, it just returns the list of LR schedulers with a BestModelRestore callback.
        expt_callbacks += self._init_lr_scheduler_callbacks(lr_schedulers)

        if callbacks is not None:
            expt_callbacks += callbacks

        try:
            return training_func(*args, initial_epoch=initial_epoch, callbacks=expt_callbacks, **kwargs)
        finally:
            if self.logging:
                self._save_history()

            if tensorboard_writer is not None:
                tensorboard_writer.close()

    def load_checkpoint(
        self, checkpoint: Union[int, str], *, verbose: bool = False, strict: bool = True
    ) -> Union[Dict, None]:
        """
        Loads the model's weights with the weights at a given checkpoint epoch.

        Args:
            checkpoint (Union[int, str]): Which checkpoint to load the model's weights form.

                - If 'best', will load the best weights according to ``monitor_metric`` and ``monitor_mode``.
                - If 'last', will load the last model checkpoint.
                - If int, will load the checkpoint of the specified epoch.
                - If a path (str), will load the model pickled state_dict weights (for instance, saved as
                  ``torch.save(a_pytorch_network.state_dict(), "./a_path.p")``).
            verbose (bool, optional): Whether or not to print the checkpoint filename, and the best epoch
                number and stats when checkpoint is 'best'.
                (Default value = False)

        Returns:
            If checkpoint is 'best', will return the best epoch stats, as per :func:`~get_best_epoch_stats()`,
            if checkpoint is 'last', will return the last epoch stats, if checkpoint is a int, will return the
            epoch number stats, if a path, will return the stats of that specific checkpoint.
            else None.
        """

        epoch_stats = None
        if isinstance(checkpoint, int):
            epoch_stats, incompatible_keys = self._load_epoch_checkpoint(checkpoint, verbose=verbose, strict=strict)
        elif checkpoint == 'best':
            epoch_stats, incompatible_keys = self._load_best_checkpoint(verbose=verbose, strict=strict)
        elif checkpoint == 'last':
            epoch_stats, incompatible_keys = self._load_last_checkpoint(verbose=verbose, strict=strict)
        else:
            incompatible_keys = self._load_path_checkpoint(path=checkpoint, verbose=verbose, strict=strict)

        if len(incompatible_keys.unexpected_keys) > 0:
            warnings.warn(
                'Unexpected key(s): ' + ', '.join(f'"{k}"' for k in incompatible_keys.unexpected_keys) + '.',
                stacklevel=2,
            )
        if len(incompatible_keys.missing_keys) > 0:
            warnings.warn(
                'Missing key(s): ' + ', '.join(f'"{k}"' for k in incompatible_keys.missing_keys) + '.',
                stacklevel=2,
            )

        return epoch_stats

    def _print_epoch_stats(self, epoch_stats):
        metrics_str = ', '.join(
            f'{metric_name}: {epoch_stats[metric_name].item():g}' for metric_name in epoch_stats.columns[2:]
        )
        print(metrics_str)

    def _load_epoch_checkpoint(self, epoch: int, *, verbose: bool = False, strict: bool = True) -> None:
        ckpt_filename = self.best_checkpoint_filename.format(epoch=epoch)

        history = self.get_stats()
        epoch_stats = history.iloc[epoch - 1 : epoch]

        if verbose:
            print(f"Loading checkpoint {ckpt_filename}")
            self._print_epoch_stats(epoch_stats)

        if not os.path.isfile(ckpt_filename):
            raise ValueError(f"No checkpoint found for epoch {epoch}")

        return epoch_stats, self.model.load_weights(ckpt_filename, strict=strict)

    def _load_best_checkpoint(self, *, verbose: bool = False, strict: bool = True) -> Dict:
        best_epoch_stats = self.get_best_epoch_stats()
        best_epoch = best_epoch_stats['epoch'].item()

        ckpt_filename = self.best_checkpoint_filename.format(epoch=best_epoch)

        if verbose:
            print(f"Found best checkpoint at epoch: {best_epoch}")
            self._print_epoch_stats(best_epoch_stats)
            print(f"Loading checkpoint {ckpt_filename}")

        if not os.path.isfile(ckpt_filename):
            raise ValueError(f"No checkpoint found for epoch {best_epoch}")

        return best_epoch_stats, self.model.load_weights(ckpt_filename, strict=strict)

    def _load_last_checkpoint(self, *, verbose: bool = False, strict: bool = True) -> None:
        history = self.get_stats()
        epoch_stats = history.iloc[-1:]

        if verbose:
            print(f"Loading checkpoint {self.model_checkpoint_filename}")
            self._print_epoch_stats(epoch_stats)

        return epoch_stats, self.model.load_weights(self.model_checkpoint_filename, strict=strict)

    def _load_path_checkpoint(self, path, verbose: bool = False, strict: bool = True) -> None:
        if verbose:
            print(f"Loading checkpoint {path}")

        return self.model.load_weights(path, strict=strict)

    def test(self, test_generator, **kwargs):
        """
        Computes and returns the loss and the metrics of the model on a given test examples
        generator.

        If the Experiment has logging enabled (i.e. self.logging is True), a checkpoint (the best one by default)
        is loaded and test and validation statistics are saved in a specific test output .tsv file. Otherwise, the
        current weights of the network is used for testing and statistics are only shown in the standard output.

        Args:
            test_generator: Generator-like object for the test set. See :func:`~Model.fit_generator()` for
                details on the types of generators supported.
            checkpoint (Union[str, int]): Which model checkpoint weights to load for the test evaluation.

                - If 'best', will load the best weights according to ``monitor_metric`` and ``monitor_mode``.
                - If 'last', will load the last model checkpoint.
                - If int, will load the checkpoint of the specified epoch.
                - If a path (str), will load the model pickled state_dict weights (for instance, saved as
                  ``torch.save(a_pytorch_network.state_dict(), "./a_path.p")``).

                This argument has no effect when logging is disabled. (Default value = 'best')
            seed (int, optional): Seed used to make the sampling deterministic.
                (Default value = 42)
            name (str): Prefix of the test log file. (Default value = 'test')
            kwargs: Any keyword arguments to pass to :func:`~Model.evaluate_generator()`.

        If the Experiment has logging enabled (i.e. self.logging is True), one callback will be automatically
        included to save the test metrics. Moreover, a :class:`~poutyne.AtomicCSVLogger` will save the test
        metrics in an output .tsv file.

        Returns:
            dict sorting of all the test metrics values by their names.
        """
        return self._test(self.model.evaluate_generator, test_generator, **kwargs)

    def test_dataset(self, test_dataset, **kwargs) -> Dict:
        """
        Computes and returns the loss and the metrics of the model on a given test dataset.

        If the Experiment has logging enabled (i.e. self.logging is True), a checkpoint (the best one by default)
        is loaded and test and validation statistics are saved in a specific test output .tsv file. Otherwise, the
        current weights of the network is used for testing and statistics are only shown in the standard output.

        Args:
            test_dataset (~torch.utils.data.Dataset): Test dataset.
            checkpoint (Union[str, int]): Which model checkpoint weights to load for the test evaluation.

                - If 'best', will load the best weights according to ``monitor_metric`` and ``monitor_mode``.
                - If 'last', will load the last model checkpoint.
                - If int, will load the checkpoint of the specified epoch.
                - If a path (str), will load the model pickled state_dict weights (for instance, saved as
                  ``torch.save(a_pytorch_network.state_dict(), "./a_path.p")``).

                This argument has no effect when logging is disabled. (Default value = 'best')
            seed (int, optional): Seed used to make the sampling deterministic.
                (Default value = 42)
            name (str): Prefix of the test log file. (Default value = 'test')
            kwargs: Any keyword arguments to pass to :func:`~Model.evaluate_dataset()`.

        If the Experiment has logging enabled (i.e. self.logging is True), one callback will be automatically
        included to save the test metrics. Moreover, a :class:`~poutyne.AtomicCSVLogger` will save the test
        metrics in an output .tsv file.

        Returns:
            dict sorting of all the test metrics values by their names.
        """
        return self._test(self.model.evaluate_dataset, test_dataset, **kwargs)

    def test_data(self, x, y, **kwargs) -> Dict:
        """
        Computes and returns the loss and the metrics of the model on a given test dataset.

        If the Experiment has logging enabled (i.e. self.logging is True), a checkpoint (the best one by default)
        is loaded and test and validation statistics are saved in a specific test output .tsv file. Otherwise, the
        current weights of the network is used for testing and statistics are only shown in the standard output.

        Args:
            x (Union[~torch.Tensor, ~numpy.ndarray] or Union[tuple, list] of Union[~torch.Tensor, ~numpy.ndarray]):
                Input to the model. Union[Tensor, ndarray] if the model has a single input.
                Union[tuple, list] of Union[Tensor, ndarray] if the model has multiple inputs.
            y (Union[~torch.Tensor, ~numpy.ndarray] or Union[tuple, list] of Union[~torch.Tensor, ~numpy.ndarray]):
                Target, corresponding ground truth.
                Union[Tensor, ndarray] if the model has a single output.
                Union[tuple, list] of Union[Tensor, ndarray] if the model has multiple outputs.
            checkpoint (Union[str, int]): Which model checkpoint weights to load for the test evaluation.

                - If 'best', will load the best weights according to ``monitor_metric`` and ``monitor_mode``.
                - If 'last', will load the last model checkpoint.
                - If int, will load the checkpoint of the specified epoch.
                - If a path (str), will load the model pickled state_dict weights (for instance, saved as
                  ``torch.save(a_pytorch_network.state_dict(), "./a_path.p")``).

                This argument has no effect when logging is disabled. (Default value = 'best')
            seed (int, optional): Seed used to make the sampling deterministic.
                (Default value = 42)
            name (str): Prefix of the test log file. (Default value = 'test')
            kwargs: Any keyword arguments to pass to :func:`~Model.evaluate()`.
        If the Experiment has logging enabled (i.e. self.logging is True), one callback will be automatically
        included to save the test metrics. Moreover, a :class:`~poutyne.AtomicCSVLogger` will save the test
        metrics in an output .tsv file.

        Returns:
            dict sorting of all the test metrics values by their names.
        """
        return self._test(self.model.evaluate, x, y, **kwargs)

    def _test(
        self,
        evaluate_func,
        *args,
        checkpoint: Union[str, int] = 'best',
        seed: int = 42,
        name='test',
        verbose=True,
        **kwargs,
    ) -> Dict:
        if kwargs.get('return_dict_format') is False:
            raise ValueError("This method only returns a dict.")
        kwargs['return_dict_format'] = True

        set_seeds(seed)

        if self.logging:
            if not self.monitoring and checkpoint == 'best':
                checkpoint = 'last'
            epoch_stats = self.load_checkpoint(checkpoint, verbose=verbose)

        if verbose:
            print(f"Running {name}")
        ret = evaluate_func(*args, **kwargs, verbose=verbose)

        if self.logging:
            test_metrics_dict = ret[0] if isinstance(ret, tuple) else ret
            test_stats = pd.DataFrame([list(test_metrics_dict.values())], columns=list(test_metrics_dict.keys()))
            test_stats.drop(['time'], axis=1, inplace=True)
            if epoch_stats is not None:
                epoch_stats = epoch_stats.reset_index(drop=True)
                test_stats = epoch_stats.join(test_stats)
            test_stats.to_csv(self.test_log_filename.format(name=name), sep='\t', index=False)

        return ret

    def infer(self, generator, **kwargs) -> Any:
        """
        Returns the predictions of the network given batches of samples ``x``, where the tensors are
        converted into Numpy arrays.

        Args:
            generator: Generator-like object for the dataset. The generator must yield a batch of
                samples. See the :func:`fit_generator()` method for details on the types of generators
                supported. This should only yield input data ``x`` and NOT the target ``y``.
            checkpoint (Union[str, int]): Which model checkpoint weights to load for the prediction.

                - If 'best', will load the best weights according to ``monitor_metric`` and ``monitor_mode``.
                - If 'last', will load the last model checkpoint.
                - If int, will load the checkpoint of the specified epoch.
                - If a path (str), will load the model pickled state_dict weights (for instance, saved as
                  ``torch.save(a_pytorch_network.state_dict(), "./a_path.p")``).

                This argument has no effect when logging is disabled. (Default value = 'best')
            kwargs: Any keyword arguments to pass to :func:`~Model.predict_generator()`.

        Returns:
            Depends on the value of ``concatenate_returns``. By default, (``concatenate_returns`` is true),
            the data structures (tensor, tuple, list, dict) returned as predictions for the batches are
            merged together. In the merge, the tensors are converted into Numpy arrays and are then
            concatenated together. If ``concatenate_returns`` is false, then a list of the predictions
            for the batches is returned with tensors converted into Numpy arrays.
        """
        return self._predict(self.model.predict_generator, generator, **kwargs)

    def infer_dataset(self, dataset, **kwargs) -> Any:
        """
        Returns the inferred predictions of the network given a dataset, where the tensors are
        converted into Numpy arrays.

        Args:
            dataset (~torch.utils.data.Dataset): Dataset. Must not return ``y``, just ``x``.
            checkpoint (Union[str, int]): Which model checkpoint weights to load for the prediction.

                - If 'best', will load the best weights according to ``monitor_metric`` and ``monitor_mode``.
                - If 'last', will load the last model checkpoint.
                - If int, will load the checkpoint of the specified epoch.
                - If a path (str), will load the model pickled state_dict weights (for instance, saved as
                  ``torch.save(a_pytorch_network.state_dict(), "./a_path.p")``).

                This argument has no effect when logging is disabled. (Default value = 'best')
            kwargs: Any keyword arguments to pass to :func:`~Model.predict_dataset()`.

        Returns:
            Return the predictions in the format outputted by the model.
        """
        return self._predict(self.model.predict_dataset, dataset, **kwargs)

    def infer_data(self, x, **kwargs) -> Any:
        """
        Returns the inferred predictions of the network given a dataset ``x``, where the tensors are
        converted into Numpy arrays.

        Args:
            x (Union[~torch.Tensor, ~numpy.ndarray] or Union[tuple, list] of Union[~torch.Tensor, ~numpy.ndarray]):
                Input to the model. Union[Tensor, ndarray] if the model has a single input.
                Union[tuple, list] of Union[Tensor, ndarray] if the model has multiple inputs.
            checkpoint (Union[str, int]): Which model checkpoint weights to load for the prediction.

                - If 'best', will load the best weights according to ``monitor_metric`` and ``monitor_mode``.
                - If 'last', will load the last model checkpoint.
                - If int, will load the checkpoint of the specified epoch.
                - If a path (str), will load the model pickled state_dict weights (for instance, saved as
                  ``torch.save(a_pytorch_network.state_dict(), "./a_path.p")``).

                This argument has no effect when logging is disabled. (Default value = 'best')
            kwargs: Any keyword arguments to pass to :func:`~Model.predict()`.

        Returns:
            Return the predictions in the format outputted by the model.
        """
        return self._predict(self.model.predict, x, **kwargs)

    def _predict(
        self, evaluate_func: Callable, *args, verbose=True, checkpoint: Union[str, int] = 'best', **kwargs
    ) -> Any:
        if self.logging:
            if not self.monitoring and checkpoint == 'best':
                checkpoint = 'last'
            self.load_checkpoint(checkpoint, verbose=verbose)

        ret = evaluate_func(*args, verbose=verbose, **kwargs)
        return ret

    def is_better_than(self, another_experiment) -> bool:
        """
        Compare the results of the Experiment with another experiment. To compare, both Experiments need to be
        logged, monitor the same metric and the same monitor mode ("min" or "max").

        Args:
            another_experiment (~poutyne. Experiment): Another Poutyne experiment to compare results with.

        Return:
            Whether the Experiment is better than the Experiment to compare with.
        """
        if not self.logging:
            raise ValueError("The experiment is not logged.")
        if not another_experiment.logging:
            raise ValueError("The experiment to compare to is not logged.")

        if self.monitor_metric != another_experiment.monitor_metric:
            raise ValueError("The monitored metric is not the same between the two experiment.")
        monitored_metric = self.monitor_metric

        if self.monitor_mode != another_experiment.monitor_mode:
            raise ValueError("The monitored mode is not the same between the two experiment.")
        monitor_mode = self.monitor_mode

        checkpoint = 'best' if self.monitoring else 'last'
        self_stats = self.load_checkpoint(checkpoint, verbose=False)
        self_monitored_metric = self_stats[monitored_metric]
        self_monitored_metric_value = self_monitored_metric.item()

        other_checkpoint = 'best' if another_experiment.monitoring else 'last'
        other_stats = self.load_checkpoint(other_checkpoint, verbose=False)
        other_monitored_metric = other_stats[monitored_metric]
        other_monitored_metric_value = other_monitored_metric.item()

        if monitor_mode == 'min':
            is_better_than = self_monitored_metric_value < other_monitored_metric_value
        else:
            is_better_than = self_monitored_metric_value > other_monitored_metric_value
        return is_better_than



import contextlib
from collections import OrderedDict
from itertools import islice, chain
from math import cos, pi
from typing import Dict, List, Tuple, Optional

try:
    import matplotlib.pyplot as plt
except ImportError:
    plt = None

from .callbacks import Callback


###############################################################################
# Lazy parameter spaces
# A space is just an iterable
class linspace:
    """
    A lazy linear parameter space that goes from ``start`` to ``end`` in ``steps`` steps.

    Args:
        start (int): the start point.
        end (int): the end point.
        steps (int): the number of steps between start and end.

    Example:
        >>> list(linspace(0, 1, 3))
        [0.0, 0.5, 1.0]
    """

    def __init__(self, start: int, end: int, steps: int):
        self.start = start
        self.end = end
        self.steps = steps

    def _progress(self, i):
        return i / (self.steps - 1)

    def __iter__(self):
        return (self.start + self._progress(i) * (self.end - self.start) for i in range(self.steps))


class cosinespace:
    """
    A lazy cosine parameter space that goes from ``start`` to ``end`` in ``steps`` steps.

    Args:
        start (int): the start point.
        end (int): the end point.
        steps (int): the number of steps between start and end.

    Example:
        >>> list(cosinespace(0, 1, 3))
        [0.0, 0.5, 1.0]
    """

    def __init__(self, start: int, end: int, steps: int):
        self.start = start
        self.end = end
        self.steps = steps

    def _progress(self, i):
        return i / (self.steps - 1)

    def __iter__(self):
        return (self.end + (self.start - self.end) * (1 + cos(self._progress(i) * pi)) / 2 for i in range(self.steps))


###############################################################################
class Phase:
    """
    Defines how to configure an optimizer.

    For each train step it returns a dictionary that contains the configuration for the optimizer.

    Args:
        lr (List[float], optional): a configuration space for the learning rate.
        momentum (List[float], optional): a configuration space for the momentum.
    """

    def __init__(self, *, lr: Optional[float] = None, momentum: Optional[float] = None):
        if lr is None and momentum is None:
            raise ValueError("You must specify lr and/or momentum.")

        self.configuration = OrderedDict()
        if lr is not None:
            self.configuration["lr"] = lr
        if momentum is not None:
            self.configuration["momentum"] = momentum

    def __iter__(self):
        names = list(self.configuration.keys())
        values = self.configuration.values()
        for values in zip(*self.configuration.values()):
            yield dict(zip(names, values))

    def __repr__(self):
        return "\n".join(
            [
                "Phase:",
                *[f"    {name}: {val}" for name, val in self.configuration.items()],
            ]
        )

    def plot(self, param_name: str = "lr", ax=None):
        """
        Plot the phase for the given `param_name`.

        Args:
            param_name (str, optional): the name of the parameter to plot.
            ax (~matplotlib.pyplot.axis, optional): a matplotlib axis to plot on, if given.

        Returns:
            The matplotlib axis.
        """
        # pylint: disable=import-error

        if ax is None:
            _fig, ax = plt.subplots()

        ax.plot(list(self.configuration[param_name]))
        ax.set_xlabel("steps")
        ax.set_ylabel(param_name)
        return ax


###############################################################################
# complex policies build from simple phases
# pylint
def one_cycle_phases(
    steps: int,
    lr: Tuple[float, float] = (0.1, 1),
    momentum: Tuple[float, float] = (0.95, 0.85),
    finetune_lr: float = 0.01,
    finetune_fraction: float = 0.1,
) -> List[Phase]:
    """
    The "one-cycle" policy as described in the paper `Super-Convergence: Very Fast Training of
    Neural Networks Using Large Learning Rates <https://arxiv.org/abs/1708.07120>`_.

    You might want to read the paper and adjust the parameters.

    Args:
        steps (int): the total number of steps to take.
        lr (Tuple[float, float]): tuple for the triangular learning rate (start, middle).
        momentum (Tuple[float, float]): tuple for the triangular momentum (start, middle).
        finetune_lr (float): target learning rate for the final fine tuning. Should be smaller than
            `min(lr)`.
        finetune_fraction (float): fraction of steps used for the fine tuning.
            Must be between 0 and 1.

    Returns:
        A list of configured :class:`~poutyne.Phase` instances.

    References:
        `Super-Convergence: Very Fast Training of Neural Networks Using Large Learning Rates
        <https://arxiv.org/abs/1708.07120>`_
    """
    steps_annealing = int(steps * finetune_fraction)
    steps_up = (steps - steps_annealing) // 2
    steps_down = steps - steps_annealing - steps_up
    return [
        Phase(
            lr=linspace(lr[0], lr[1], steps_up),
            momentum=linspace(momentum[0], momentum[1], steps_up),
        ),
        Phase(
            lr=linspace(lr[1], lr[0], steps_down),
            momentum=linspace(momentum[1], momentum[0], steps_down),
        ),
        Phase(
            lr=linspace(lr[0], finetune_lr, steps_annealing),
            momentum=linspace(momentum[0], momentum[0], steps_annealing),
        ),
    ]


def sgdr_phases(
    base_cycle_length: int,
    cycles: int,
    lr: Tuple[float, float] = (1.0, 0.1),
    cycle_mult: int = 2,
) -> List[Phase]:
    """
    The "SGDR" policy as described in the paper `SGDR: Stochastic Gradient Descent with Warm Restarts
    <https://arxiv.org/abs/1608.03983>`_.

    Note the total number of steps is calculated like this: `total_steps = sum(base_cycle_length *
    (cycle_mult ** i) for i in range(cycles))`

    You might want to read the paper and adjust the parameters.

    Args:
        base_cycle_length (int): number of steps for the first cycle.
        cycles (int): the number of repetitions.
        lr (Typle[float, float]): tuple for the learning rate for one cycle: (start, end).
        cycle_mult (float): multiply the last cycle length with this every cycle. The length of a cycle
            grows exponentially.

    Returns:
        A list of configured :class:`~poutyne.Phase` instances.

    References:
        `SGDR: Stochastic Gradient Descent with Warm Restarts
        <https://arxiv.org/abs/1608.03983>`_
    """
    steps = [base_cycle_length * (cycle_mult ** i) for i in range(cycles)]
    return [Phase(lr=cosinespace(lr[0], lr[1], step)) for step in steps]


###############################################################################
class OptimizerPolicy(Callback):
    """
    Combine different :class:`~poutyne.Phase` instances
    in an :class:`~poutyne.OptimizerPolicy` and execute the policies in a
    row.

    Args:
        phases (List[~poutyne.Phase]):
            A list of :class:`~poutyne.Phase` instances.
        initial_step (int): The step to start the policy in. Used for restarting.
    """

    def __init__(self, phases: List, *, initial_step: int = 0):
        super().__init__()
        self.phases = phases
        self.current_step = initial_step
        self.phases_iter = iter(self)

    def on_train_batch_begin(self, batch_number: int, logs: Dict):
        # Don't do anything when we run out of phases.
        with contextlib.suppress(StopIteration):
            spec = next(self.phases_iter)
            self._update_optimizer(spec)

    def __iter__(self):
        space_iter = islice(chain.from_iterable(self.phases), self.current_step, None)
        for param_dict in space_iter:
            self.current_step += 1
            yield param_dict

    def all_steps(self) -> List[Dict]:
        """
        Return the list of dictionaries of configurations for all steps.

        This does not advance the current_step count.

        Returns:
            A list of dictionaries of all the parameters for each step.
        """
        return chain.from_iterable(self.phases)

    def __repr__(self):
        return f"OptimizerPolicy:\n    phases: {self.current_step}\n    current_step: {len(self.phases)}"

    def _update_optimizer(self, param_dict: Dict):
        for param_name, param_value in param_dict.items():
            for group in self.model.optimizer.param_groups:
                group[param_name] = param_value

    def plot(self, param_name: str = "lr", ax=None):
        """
        Visualize all  :class:`~poutyne.Phase`s of
        this  :class:`~poutyne.OptimizerPolicy`.

        Args:
            param_name (str, optional): the name of the parameter to plot.
            ax (~matplotlib.pyplot.axis): a matplotlib axis to plot on, if given.

        Returns:
            The matplotlib axis.
        """
        # pylint: disable=import-error
        if ax is None:
            _fig, ax = plt.subplots()

        values = [step[param_name] for step in self.all_steps()]
        ax.plot(values)
        ax.set_ylabel(param_name)
        ax.set_xlabel("steps")
        return ax



import os
from typing import Dict, IO

import unittest
from unittest import TestCase

from tempfile import TemporaryDirectory

import torch
import torch.nn as nn

from poutyne import Model, PeriodicSaveCallback
from tests.framework.tools import some_data_generator


class PeriodicEpochSave(PeriodicSaveCallback):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, open_mode='w', read_mode='r', **kwargs)
        self.epoch_number = None

    def save_file(self, fd: IO, epoch_number: int, logs: Dict):
        print(epoch_number, file=fd)
        self.last_saved_epoch_number = epoch_number

    def restore(self, fd: IO):
        self.restored_epoch_number = int(fd.read())


class PeriodicSaveTest(TestCase):
    batch_size = 20

    def setUp(self):
        torch.manual_seed(42)
        self.pytorch_network = nn.Linear(1, 1)
        self.loss_function = nn.MSELoss()
        self.optimizer = torch.optim.SGD(self.pytorch_network.parameters(), lr=1e-3)
        self.model = Model(self.pytorch_network, self.optimizer, self.loss_function)
        self.temp_dir_obj = TemporaryDirectory()
        self.save_filename = os.path.join(self.temp_dir_obj.name, 'my_checkpoint_{epoch}.ckpt')

    def tearDown(self):
        self.temp_dir_obj.cleanup()

    def test_integration(self):
        train_gen = some_data_generator(PeriodicSaveTest.batch_size)
        valid_gen = some_data_generator(PeriodicSaveTest.batch_size)
        saver = PeriodicEpochSave(self.save_filename, monitor='val_loss', verbose=True, save_best_only=True)
        self.model.fit_generator(train_gen, valid_gen, epochs=10, steps_per_epoch=5, callbacks=[saver])

    def test_integration_with_keep_only_last_best(self):
        train_gen = some_data_generator(PeriodicSaveTest.batch_size)
        valid_gen = some_data_generator(PeriodicSaveTest.batch_size)
        saver = PeriodicEpochSave(
            self.save_filename, monitor='val_loss', verbose=True, save_best_only=True, keep_only_last_best=True
        )
        self.model.fit_generator(train_gen, valid_gen, epochs=10, steps_per_epoch=5, callbacks=[saver])

    def test_temporary_filename_arg(self):
        tmp_filename = os.path.join(self.temp_dir_obj.name, 'my_checkpoint.tmp.ckpt')
        save_filename = os.path.join(self.temp_dir_obj.name, 'my_checkpoint.ckpt')
        train_gen = some_data_generator(PeriodicSaveTest.batch_size)
        valid_gen = some_data_generator(PeriodicSaveTest.batch_size)
        saver = PeriodicEpochSave(
            save_filename, monitor='val_loss', verbose=True, period=1, temporary_filename=tmp_filename
        )
        self.model.fit_generator(train_gen, valid_gen, epochs=10, steps_per_epoch=5, callbacks=[saver])
        self.assertFalse(os.path.isfile(tmp_filename))
        self.assertTrue(os.path.isfile(save_filename))

    def test_temporary_filename_arg_with_differing_save_filename(self):
        epochs = 10
        tmp_filename = os.path.join(self.temp_dir_obj.name, 'my_checkpoint.tmp.ckpt')
        save_filename = os.path.join(self.temp_dir_obj.name, 'my_checkpoint_{epoch}.ckpt')
        train_gen = some_data_generator(PeriodicSaveTest.batch_size)
        valid_gen = some_data_generator(PeriodicSaveTest.batch_size)
        saver = PeriodicEpochSave(
            save_filename, monitor='val_loss', verbose=True, period=1, temporary_filename=tmp_filename
        )
        self.model.fit_generator(train_gen, valid_gen, epochs=epochs, steps_per_epoch=5, callbacks=[saver])
        self.assertFalse(os.path.isfile(tmp_filename))
        for i in range(1, epochs + 1):
            self.assertTrue(os.path.isfile(save_filename.format(epoch=i)))

    def test_non_atomic_write(self):
        save_filename = os.path.join(self.temp_dir_obj.name, 'my_checkpoint.ckpt')
        train_gen = some_data_generator(PeriodicSaveTest.batch_size)
        valid_gen = some_data_generator(PeriodicSaveTest.batch_size)
        saver = PeriodicEpochSave(save_filename, monitor='val_loss', verbose=True, period=1, atomic_write=False)
        self.model.fit_generator(train_gen, valid_gen, epochs=10, steps_per_epoch=5, callbacks=[saver])
        self.assertTrue(os.path.isfile(save_filename))

    def test_save_best_only(self):
        saver = PeriodicEpochSave(self.save_filename, monitor='val_loss', verbose=True, save_best_only=True)

        val_losses = [10, 3, 8, 5, 2]
        has_checkpoints = [True, True, False, False, True]
        self._test_saver_with_val_losses(saver, val_losses, has_checkpoints)

    def test_save_best_only_with_keep_only_last_best(self):
        saver = PeriodicEpochSave(
            self.save_filename, monitor='val_loss', verbose=True, save_best_only=True, keep_only_last_best=True
        )

        val_losses = [10, 3, 8, 5, 2]
        has_checkpoints = [True, True, False, False, True]
        self._test_saver_with_val_losses(saver, val_losses, has_checkpoints, keep_only_last_best=True)

    def test_save_best_only_with_max(self):
        saver = PeriodicEpochSave(self.save_filename, monitor='val_loss', mode='max', verbose=True, save_best_only=True)

        val_losses = [2, 3, 8, 5, 2]
        has_checkpoints = [True, True, True, False, False]
        self._test_saver_with_val_losses(saver, val_losses, has_checkpoints)

    def test_save_best_only_with_max_and_keep_only_last_best(self):
        saver = PeriodicEpochSave(
            self.save_filename,
            monitor='val_loss',
            mode='max',
            verbose=True,
            save_best_only=True,
            keep_only_last_best=True,
        )

        val_losses = [2, 3, 8, 5, 2]
        has_checkpoints = [True, True, True, False, False]
        self._test_saver_with_val_losses(saver, val_losses, has_checkpoints, keep_only_last_best=True)

    def test_periodic_with_period_of_1(self):
        saver = PeriodicEpochSave(self.save_filename, monitor='val_loss', verbose=True, period=1, save_best_only=False)

        val_losses = [1] * 10
        has_checkpoints = [True] * 10
        self._test_saver_with_val_losses(saver, val_losses, has_checkpoints)

    def test_periodic_with_period_of_2(self):
        saver = PeriodicEpochSave(self.save_filename, monitor='val_loss', verbose=True, period=2, save_best_only=False)

        val_losses = [1] * 10
        has_checkpoints = [False, True] * 5
        self._test_saver_with_val_losses(saver, val_losses, has_checkpoints)

    def test_keep_only_last_best_without_save_best_only(self):
        with self.assertRaises(ValueError):
            PeriodicEpochSave(
                self.save_filename, monitor='val_loss', verbose=True, save_best_only=False, keep_only_last_best=True
            )

        with self.assertRaises(ValueError):
            PeriodicEpochSave(self.save_filename, monitor='val_loss', verbose=True, keep_only_last_best=True)

    def test_save_best_only_with_restore_best(self):
        checkpointer = PeriodicEpochSave(
            self.save_filename, monitor='val_loss', verbose=True, save_best_only=True, restore_best=True
        )

        val_losses = [10, 3, 8, 7, 2, 5]
        has_checkpoints = [True, True, False, False, True, False]
        self._test_saver_with_val_losses(checkpointer, val_losses, has_checkpoints)

        self.assertEqual(5, checkpointer.restored_epoch_number)
        self.assertEqual(5, checkpointer.last_saved_epoch_number)

    def test_restore_best_without_save_best_only(self):
        with self.assertRaises(ValueError):
            PeriodicEpochSave(
                self.save_filename, monitor='val_loss', verbose=True, save_best_only=False, restore_best=True
            )

        with self.assertRaises(ValueError):
            PeriodicEpochSave(self.save_filename, monitor='val_loss', verbose=True, restore_best=True)

    def _test_saver_with_val_losses(self, saver, val_losses, has_checkpoints, keep_only_last_best=False):
        generator = some_data_generator(PeriodicSaveTest.batch_size)

        best_checkpoint_filenames = []
        saver.set_params({'epochs': len(val_losses), 'steps': 1})
        saver.set_model(self.model)
        saver.on_train_begin({})
        for epoch, (val_loss, has_checkpoint) in enumerate(zip(val_losses, has_checkpoints), 1):
            saver.on_epoch_begin(epoch, {})
            saver.on_train_batch_begin(1, {})
            loss = self._update_model(generator)
            saver.on_train_batch_end(1, {'batch': 1, 'size': PeriodicSaveTest.batch_size, 'loss': loss})
            saver.on_epoch_end(epoch, {'epoch': epoch, 'loss': loss, 'val_loss': val_loss})

            filename = self.save_filename.format(epoch=epoch)
            self.assertEqual(has_checkpoint, os.path.isfile(filename))
            if has_checkpoint:
                self.assertEqual(f'{epoch}\n', open(filename, 'r', encoding='utf-8').read())
                best_checkpoint_filenames.append(os.path.realpath(filename))

        files = [os.path.realpath(os.path.join(self.temp_dir_obj.name, f)) for f in os.listdir(self.temp_dir_obj.name)]
        if keep_only_last_best:
            self.assertEqual(1, len(files))
            self.assertEqual(files[0], best_checkpoint_filenames[-1])
        else:
            best_checkpoint_filenames = set(best_checkpoint_filenames)
            self.assertEqual(len(best_checkpoint_filenames), len(files))
            self.assertEqual(best_checkpoint_filenames, set(files))

        saver.on_train_end({})

    def _update_model(self, generator):
        self.pytorch_network.zero_grad()

        x, y = next(generator)
        pred_y = self.pytorch_network(x)
        loss = self.loss_function(pred_y, y)
        loss.backward()

        self.optimizer.step()

        return float(loss)


if __name__ == '__main__':
    unittest.main()



from typing import List, Union
from unittest import TestCase
from unittest.mock import patch, MagicMock, call

from poutyne import ProgressionCallback


class ProgressTest(TestCase):
    def setUp(self) -> None:
        self.epochs = 2
        self.num_steps = 10
        self.num_valid_steps = 10
        self.step_times_rate = 1.0
        self.any_log = {"time": self.step_times_rate, "loss": 1.0}
        self.total_time = 5
        self.final_log = {"time": self.total_time, "loss": 1.0}
        self.an_empty_log = {}

        self.metrics_str = ''

        self.coloring = True

    def test_givenATrainWithStepsValidWithoutStepsLoop_whenAlternatingLoop_thenProgressBarIsProperlySet(self):
        params = {'epochs': 2, 'steps': self.num_steps, 'valid_steps': None}

        with patch("poutyne.framework.progress.ColorProgress") as color_progress_patch:
            self.progression_callback = ProgressionCallback()
            self.progression_callback.set_params(params)
            self.progression_callback.set_model(MagicMock())

            self.progression_callback.on_train_begin(self.an_empty_log)
            color_progress_calls = [call(coloring=self.coloring)]

            # first train loop
            epoch = 1
            color_progress_calls.extend(self._a_training_loop(epoch, self.num_steps, with_progress_bar=True))

            # first valid loop
            color_progress_calls.extend(self._a_valid_loop(epoch, self.num_steps, with_progress_bar=False))

            # second train loop
            epoch = 2
            color_progress_calls.extend(self._a_training_loop(epoch, self.num_steps, with_progress_bar=True))

            # second valid loop
            color_progress_calls.extend(self._a_valid_loop(epoch, self.num_steps, with_progress_bar=False))

            color_progress_patch.assert_has_calls(color_progress_calls)

    def test_givenATrainWithoutStepsValidWitStepsLoop_whenAlternatingLoop_thenProgressBarIsProperlySet(self):
        params = {'epochs': 2, 'steps': None, 'valid_steps': self.num_valid_steps}

        with patch("poutyne.framework.progress.ColorProgress") as color_progress_patch:
            self.progression_callback = ProgressionCallback()
            self.progression_callback.set_params(params)
            self.progression_callback.set_model(MagicMock())

            self.progression_callback.on_train_begin(self.an_empty_log)
            color_progress_calls = [call(coloring=self.coloring)]

            # first train loop
            epoch = 1
            color_progress_calls.extend(self._a_training_loop(epoch, self.num_steps, with_progress_bar=False))

            # first valid loop
            color_progress_calls.extend(self._a_valid_loop(epoch, self.num_steps, with_progress_bar=True))

            # second train loop
            epoch = 2
            color_progress_calls.extend(self._a_training_loop(epoch, self.num_steps, with_progress_bar=False))

            # second valid loop
            color_progress_calls.extend(self._a_valid_loop(epoch, self.num_steps, with_progress_bar=True))

            color_progress_patch.assert_has_calls(color_progress_calls)

    def test_givenATrainWithoutStepsValidWithoutStepsLoop_whenAlternatingLoop_thenProgressBarIsProperlySet(self):
        params = {'epochs': 2, 'steps': None, 'valid_steps': None}

        with patch("poutyne.framework.progress.ColorProgress") as color_progress_patch:
            self.progression_callback = ProgressionCallback()
            self.progression_callback.set_params(params)
            self.progression_callback.set_model(MagicMock())

            self.progression_callback.on_train_begin(self.an_empty_log)
            color_progress_calls = [call(coloring=self.coloring)]

            # first train loop
            epoch = 1
            color_progress_calls.extend(self._a_training_loop(epoch, self.num_steps, with_progress_bar=False))

            # first valid loop
            color_progress_calls.extend(self._a_valid_loop(epoch, self.num_steps, with_progress_bar=False))

            # second train loop
            epoch = 2
            color_progress_calls.extend(self._a_training_loop(epoch, self.num_steps, with_progress_bar=False))

            # second valid loop
            color_progress_calls.extend(self._a_valid_loop(epoch, self.num_steps, with_progress_bar=False))

            color_progress_patch.assert_has_calls(color_progress_calls)

    def test_givenATrainWithTrainStepsAndValidStepsLoops_thenProgressBarIsProperlySet(self):
        params = {'epochs': 2, 'steps': self.num_steps, 'valid_steps': self.num_valid_steps}

        with patch("poutyne.framework.progress.ColorProgress") as color_progress_patch:
            self.progression_callback = ProgressionCallback()
            self.progression_callback.set_params(params)
            self.progression_callback.set_model(MagicMock())

            self.progression_callback.on_train_begin(self.an_empty_log)
            color_progress_calls = [call(coloring=self.coloring)]

            # first train loop
            epoch = 1
            color_progress_calls.extend(self._a_training_loop(epoch, self.num_steps, with_progress_bar=True))

            # first valid loop
            color_progress_calls.extend(self._a_valid_loop(epoch, self.num_steps, with_progress_bar=True))

            # second train loop
            epoch = 2
            color_progress_calls.extend(self._a_training_loop(epoch, self.num_steps, with_progress_bar=True))

            # second valid loop
            color_progress_calls.extend(self._a_valid_loop(epoch, self.num_steps, with_progress_bar=True))

            color_progress_patch.assert_has_calls(color_progress_calls)

    def test_givenATrainLoop_whenShowingEveryNSteps_thenProgressBarIsProperlySet(self):
        params = {'epochs': 2, 'steps': self.num_steps, 'valid_steps': self.num_valid_steps}

        with patch("poutyne.framework.progress.ColorProgress") as color_progress_patch:
            self.progression_callback = ProgressionCallback(show_every_n_train_steps=2, show_every_n_valid_steps=3)
            self.progression_callback.set_params(params)
            self.progression_callback.set_model(MagicMock())

            self.progression_callback.on_train_begin(self.an_empty_log)
            color_progress_calls = [call(coloring=self.coloring)]

            # first train loop
            epoch = 1
            color_progress_calls.extend(
                self._a_training_loop(epoch, self.num_steps, with_progress_bar=True, show_every_n_steps=2)
            )

            # first valid loop
            color_progress_calls.extend(
                self._a_valid_loop(epoch, self.num_steps, with_progress_bar=True, show_every_n_steps=3)
            )

            # second train loop
            epoch = 2
            color_progress_calls.extend(
                self._a_training_loop(epoch, self.num_steps, with_progress_bar=True, show_every_n_steps=2)
            )

            # second valid loop
            color_progress_calls.extend(
                self._a_valid_loop(epoch, self.num_steps, with_progress_bar=True, show_every_n_steps=3)
            )

            color_progress_patch.assert_has_calls(color_progress_calls)

    def _a_training_loop(
        self,
        epoch_number,
        num_steps: int = 5,
        with_progress_bar: bool = True,
        show_every_n_steps: Union[str, int] = 'all',
    ) -> List:
        color_progress_calls = []

        self.progression_callback.on_epoch_begin(epoch_number, self.an_empty_log)
        if with_progress_bar:
            color_progress_calls.append(call().set_progress_bar(self.num_steps))
        else:
            color_progress_calls.append(call().close_progress_bar())
        color_progress_calls.append(call().on_epoch_begin(epoch_number=epoch_number, epochs=self.epochs))

        step_times_weighted_sum = 0.0
        for batch_number in range(1, num_steps + 1):
            do_print = show_every_n_steps == 'all' or (
                isinstance(show_every_n_steps, int) and batch_number % show_every_n_steps == 0
            )
            self.progression_callback.on_train_batch_end(batch_number, self.any_log)
            if with_progress_bar:
                remaining_time = self.step_times_rate * (self.num_steps - batch_number)

                color_progress_calls.append(
                    call().on_train_batch_end(
                        remaining_time=remaining_time,
                        batch_number=batch_number,
                        metrics_str=self.metrics_str,
                        steps=self.num_steps,
                        do_print=do_print,
                    )
                )
            else:
                step_times_weighted_sum += batch_number * self.step_times_rate
                normalizing_factor = (batch_number * (batch_number + 1)) / 2
                step_times_rate = step_times_weighted_sum / normalizing_factor

                color_progress_calls.append(
                    call().on_train_batch_end(
                        remaining_time=step_times_rate,
                        batch_number=batch_number,
                        metrics_str=self.metrics_str,
                        do_print=do_print,
                    )
                )
        return color_progress_calls

    def _a_valid_loop(
        self,
        epoch_number,
        num_steps: int = 5,
        with_progress_bar: bool = True,
        show_every_n_steps: Union[str, int] = 'all',
    ) -> List:
        color_progress_calls = []

        self.progression_callback.on_valid_begin(self.an_empty_log)
        if with_progress_bar:
            color_progress_calls.append(call().set_progress_bar(self.num_steps))
        else:
            color_progress_calls.append(call().close_progress_bar())
        color_progress_calls.append(call().on_valid_begin())

        step_times_weighted_sum = 0.0
        for valid_batch_number in range(1, num_steps + 1):
            do_print = show_every_n_steps == 'all' or (
                isinstance(show_every_n_steps, int) and valid_batch_number % show_every_n_steps == 0
            )
            self.progression_callback.on_valid_batch_end(valid_batch_number, self.any_log)

            if with_progress_bar:
                remaining_time = self.step_times_rate * (self.num_steps - valid_batch_number)

                color_progress_calls.append(
                    call().on_valid_batch_end(
                        remaining_time=remaining_time,
                        batch_number=valid_batch_number,
                        metrics_str=self.metrics_str,
                        steps=self.num_valid_steps,
                        do_print=do_print,
                    )
                )
            else:
                step_times_weighted_sum += valid_batch_number * self.step_times_rate
                normalizing_factor = (valid_batch_number * (valid_batch_number + 1)) / 2
                step_times_rate = step_times_weighted_sum / normalizing_factor

                color_progress_calls.append(
                    call().on_valid_batch_end(
                        remaining_time=step_times_rate,
                        batch_number=valid_batch_number,
                        metrics_str=self.metrics_str,
                        do_print=do_print,
                    )
                )

        self.progression_callback.on_valid_end(self.final_log)
        self.progression_callback.on_epoch_end(epoch_number, self.final_log)

        color_progress_calls.append(
            call().on_epoch_end(
                total_time=self.total_time,
                train_last_steps=self.num_steps,
                valid_last_steps=self.num_valid_steps,
                metrics_str=self.metrics_str,
            )
        )

        return color_progress_calls



# Because nn.Module has the abstract method _forward_unimplemented
# pylint: disable=abstract-method
import torch.nn as nn

from .metrics_registering import register_batch_metric, register_batch_metric_function


class BatchMetric(nn.Module):
    def __init__(self, reduction: str = 'mean'):
        super().__init__()
        REDUCTIONS = ['none', 'mean', 'sum']
        if reduction not in REDUCTIONS:
            raise ValueError(f"Reduction is not in {REDUCTIONS}")

        self.reduction = reduction


class Accuracy(BatchMetric):
    r"""
    This metric computes the accuracy using a similar interface to
    :class:`~torch.nn.CrossEntropyLoss`.

    Args:
        ignore_index (int): Specifies a target value that is ignored and does not contribute
            to the accuracy. (Default value = -100)
        reduction (string, optional): Specifies the reduction to apply to the output:
            ``'none'`` | ``'mean'`` | ``'sum'``. ``'none'``: no reduction will be applied,
            ``'mean'``: the sum of the output will be divided by the number of
            elements in the output, ``'sum'``: the output will be summed.


    Possible string name in :class:`batch_metrics argument <poutyne.Model>`:
        - ``'acc'``
        - ``'accuracy'``

    Keys in :class:`logs<poutyne.Callback>` dictionary of callbacks:
        - Train: ``'acc'``
        - Validation: ``'val_acc'``

    Shape:
        - Input: :math:`(N, C)` where `C = number of classes`, or
          :math:`(N, C, d_1, d_2, ..., d_K)` with :math:`K \geq 1` in the case of
          `K`-dimensional accuracy.
        - Target: :math:`(N)` where each value is :math:`0 \leq \text{targets}[i] \leq C-1`, or
          :math:`(N, d_1, d_2, ..., d_K)` with :math:`K \geq 1` in the case of
          K-dimensional accuracy.
        - Output: The accuracy.
    """

    def __init__(self, *, ignore_index: int = -100, reduction: str = 'mean'):
        super().__init__(reduction)
        self.__name__ = 'acc'
        self.ignore_index = ignore_index

    def forward(self, y_pred, y_true):
        return acc(y_pred, y_true, ignore_index=self.ignore_index, reduction=self.reduction)


@register_batch_metric('acc', 'accuracy')
def acc(y_pred, y_true, *, ignore_index=-100, reduction='mean'):
    """
    Computes the accuracy.

    This is a functional version of :class:`~poutyne.Accuracy`.

    See :class:`~poutyne.Accuracy` for details.
    """
    y_pred = y_pred.argmax(1)
    weights = (y_true != ignore_index).float()
    num_labels = weights.sum()
    acc_pred = (y_pred == y_true).float() * weights

    if reduction in ['mean', 'sum']:
        acc_pred = acc_pred.sum()

    if reduction == 'mean':
        acc_pred = acc_pred / num_labels

    return acc_pred * 100


class BinaryAccuracy(BatchMetric):
    r"""
    This metric computes the accuracy using a similar interface to
    :class:`~torch.nn.BCEWithLogitsLoss`.

    Args:
        threshold (float): the threshold for class :math:`1`. Default value is ``0.``, that is a
            probability of ``sigmoid(0.) = 0.5``.
        reduction (string, optional): Specifies the reduction to apply to the output:
            ``'none'`` | ``'mean'`` | ``'sum'``. ``'none'``: no reduction will be applied,
            ``'mean'``: the sum of the output will be divided by the number of
            elements in the output, ``'sum'``: the output will be summed.

    Possible string name in :class:`batch_metrics argument <poutyne.Model>`:
        - ``'bin_acc'``
        - ``'binary_acc'``
        - ``'binary_accuracy'``

    Keys in :class:`logs<poutyne.Callback>` dictionary of callbacks:
        - Train: ``'bin_acc'``
        - Validation: ``'val_bin_acc'``

    Shape:
        - Input: :math:`(N, *)` where :math:`*` means, any number of additional
          dimensions
        - Target: :math:`(N, *)`, same shape as the input
        - Output: The binary accuracy.
    """

    def __init__(self, *, threshold: float = 0.0, reduction: str = 'mean'):
        super().__init__(reduction)
        self.__name__ = 'bin_acc'
        self.threshold = threshold

    def forward(self, y_pred, y_true):
        return bin_acc(y_pred, y_true, threshold=self.threshold, reduction=self.reduction)


@register_batch_metric('binacc', 'binaryacc', 'binaryaccuracy')
def bin_acc(y_pred, y_true, *, threshold=0.0, reduction='mean'):
    """
    Computes the binary accuracy.

    This is a functional version of :class:`~poutyne.BinaryAccuracy`.

    See :class:`~poutyne.BinaryAccuracy` for details.
    """
    y_pred = (y_pred > threshold).float()
    acc_pred = (y_pred == y_true).float()
    if reduction == 'mean':
        acc_pred = acc_pred.mean()
    elif reduction == 'sum':
        acc_pred = acc_pred.sum()
    return acc_pred * 100


class TopKAccuracy(BatchMetric):
    r"""
    This metric computes the top-k accuracy using a similar interface to
    :class:`~torch.nn.CrossEntropyLoss`.

    Args:
        k (int): Specifies the value of ``k`` in the top-k accuracy.
        ignore_index (int): Specifies a target value that is ignored and does not contribute
            to the top-k accuracy. (Default value = -100)
        reduction (string, optional): Specifies the reduction to apply to the output:
            ``'none'`` | ``'mean'`` | ``'sum'``. ``'none'``: no reduction will be applied,
            ``'mean'``: the sum of the output will be divided by the number of
            elements in the output, ``'sum'``: the output will be summed.


    Possible string name in :class:`batch_metrics argument <poutyne.Model>`:
        - ``'top{k}'``
        - ``'top{k}_acc'``
        - ``'top{k}_accuracy'``

        for ``{k}`` from 1 to 10, 20, 30, ..., 100.

    Keys in :class:`logs<poutyne.Callback>` dictionary of callbacks:
        - Train: ``'top{k}'``
        - Validation: ``'val_top{k}'``

        where ``{k}`` is replaced by the value of parameter ``k``.

    Shape:
        - Input: :math:`(N, C)` where `C = number of classes`, or
          :math:`(N, C, d_1, d_2, ..., d_K)` with :math:`K \geq 1` in the case of
          `K`-dimensional top-k accuracy.
        - Target: :math:`(N)` where each value is :math:`0 \leq \text{targets}[i] \leq C-1`, or
          :math:`(N, d_1, d_2, ..., d_K)` with :math:`K \geq 1` in the case of
          K-dimensional top-k accuracy.
        - Output: The top-k accuracy.
    """

    def __init__(self, k: int, *, ignore_index: int = -100, reduction: str = 'mean'):
        super().__init__(reduction)
        self.__name__ = f'top{k}'
        self.k = k
        self.ignore_index = ignore_index

    def forward(self, y_pred, y_true):
        return topk(y_pred, y_true, self.k, ignore_index=self.ignore_index, reduction=self.reduction)


def topk(y_pred, y_true, k, *, ignore_index=-100, reduction='mean'):
    """
    Computes the top-k accuracy.

    This is a functional version of :class:`~poutyne.TopKAccuracy`.

    See :class:`~poutyne.TopKAccuracy` for details.
    """
    topk_pred = y_pred.topk(k, dim=1)[1]
    weights = (y_true != ignore_index).float()
    num_labels = weights.sum()
    topk_acc = (y_true.unsqueeze(1) == topk_pred).any(1).float() * weights

    if reduction in ['mean', 'sum']:
        topk_acc = topk_acc.sum()

    if reduction == 'mean':
        topk_acc = topk_acc / num_labels

    return topk_acc * 100


@register_batch_metric('top1', 'top1acc', 'top1accuracy')
def top1(y_pred, y_true, **kwargs):
    return acc(y_pred, y_true, **kwargs)


for k_value in range(2, 11):
    register_batch_metric_function(
        TopKAccuracy(k_value), [f'top{k_value}', f'top{k_value}acc', f'top{k_value}accuracy']
    )
del k_value

for k_value in range(20, 110, 10):
    register_batch_metric_function(
        TopKAccuracy(k_value), [f'top{k_value}', f'top{k_value}acc', f'top{k_value}accuracy']
    )
del k_value



from unittest import TestCase, skipIf

import numpy as np
import torch

from poutyne import Accuracy, BinaryAccuracy, TopKAccuracy, acc, bin_acc, topk


class AccuracyTest(TestCase):
    def setUp(self):
        self.predictions = torch.Tensor(
            [
                [0.35, 0.25, 0.1, 0.1, 0.2],
                [0.1, 0.6, 0.1, 0.2, 0.0],
                [0.1, 0.6, 0.1, 0.2, 0.0],
                [0.1, 0.5, 0.1, 0.2, 0.0],
                [0.1, 0.2, 0.1, 0.7, 0.0],
                [0.1, 0.6, 0.1, 0.2, 0.0],
            ]
        )
        self.label_predictions = torch.Tensor([0, 1, 1, 1, 3, 1])
        self.targets = torch.Tensor([0, 4, 1, 0, 3, 0])
        self.accuracy_none = (self.label_predictions == self.targets).float() * 100
        self.accuracy = self.accuracy_none.mean()

    def test_standard(self):
        accuracy = Accuracy()
        actual = accuracy(self.predictions, self.targets)
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal(self.accuracy, actual)

    def test_functional(self):
        actual = acc(self.predictions, self.targets)
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal(self.accuracy, actual)

    def test_hundred(self):
        accuracy = Accuracy()
        actual = accuracy(self.predictions, self.label_predictions)
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal(100.0, actual)

    def test_sum(self):
        accuracy = Accuracy(reduction='sum')
        actual = accuracy(self.predictions, self.targets)
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal(self.accuracy * len(self.predictions), actual)

    def test_functional_sum(self):
        actual = acc(self.predictions, self.targets, reduction='sum')
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal(self.accuracy * len(self.predictions), actual)

    def test_sum_hundred(self):
        accuracy = Accuracy(reduction='sum')
        actual = accuracy(self.predictions, self.label_predictions)
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal(100.0 * len(self.predictions), actual)

    def test_none(self):
        accuracy = Accuracy(reduction='none')
        actual = accuracy(self.predictions, self.targets)
        self.assertEqual(self.targets.shape, actual.shape)
        self.assertTrue(torch.all(self.accuracy_none == actual))

    def test_functional_none(self):
        actual = acc(self.predictions, self.targets, reduction='none')
        self.assertEqual(self.targets.shape, actual.shape)
        self.assertTrue(torch.all(self.accuracy_none == actual))

    def test_none_hundred(self):
        accuracy = Accuracy(reduction='none')
        actual = accuracy(self.predictions, self.label_predictions)
        self.assertEqual(self.targets.shape, actual.shape)
        self.assertTrue(torch.all(100.0 * torch.ones_like(self.label_predictions).float() == actual))

    @skipIf(not torch.cuda.is_available(), "no gpu available")
    def test_on_gpu(self):
        accuracy = Accuracy()
        accuracy.cuda()
        actual = accuracy(self.predictions.cuda(), self.targets.cuda())
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal(self.accuracy, actual.cpu())

    @skipIf(not torch.cuda.is_available(), "no gpu available")
    def test_functional_on_gpu(self):
        actual = acc(self.predictions.cuda(), self.targets.cuda())
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal(self.accuracy, actual.cpu())


class IgnoreIndexAccuracyTest(TestCase):
    def setUp(self):
        self.predictions = torch.Tensor(
            [
                [0.35, 0.25, 0.1, 0.1, 0.2],
                [0.1, 0.6, 0.1, 0.2, 0.0],
                [0.1, 0.6, 0.1, 0.2, 0.0],
                [0.1, 0.5, 0.1, 0.2, 0.0],
                [0.1, 0.2, 0.1, 0.7, 0.0],
                [0.1, 0.6, 0.1, 0.2, 0.0],
            ]
        )
        self.label_predictions = torch.Tensor([0, 1, 1, 1, 3, 1])
        self.targets = torch.Tensor([-100, 4, 1, -100, 3, 0])
        self.accuracy = 50.0

    def test_standard(self):
        accuracy = Accuracy()
        actual = accuracy(self.predictions, self.targets)
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal(self.accuracy, actual)

    def test_functional_standard(self):
        actual = acc(self.predictions, self.targets)
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal(self.accuracy, actual)

    def test_ignore_index_with_different_value(self):
        accuracy = Accuracy(ignore_index=-1)
        actual = accuracy(self.predictions, self.targets)
        self.assertEqual((), actual.shape)
        self.assertAlmostEqual(1 / 3 * 100, float(actual), places=5)

    def test_functional_ignore_index_with_different_value(self):
        actual = acc(self.predictions, self.targets, ignore_index=-1)
        self.assertEqual((), actual.shape)
        self.assertAlmostEqual(1 / 3 * 100, float(actual), places=5)

    def test_ignore_index_with_valid_index(self):
        accuracy = Accuracy(ignore_index=1)
        actual = accuracy(self.predictions, self.targets)
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal(20.0, actual)

    def test_functional_ignore_index_with_valid_index(self):
        actual = acc(self.predictions, self.targets, ignore_index=1)
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal(20.0, actual)


class BinaryAccuracyTest(TestCase):
    def setUp(self):
        self.predictions = torch.Tensor([-0.45, -0.66, 0.8, -1.65, 0.42, -0.04, -0.99, -0.46, 1.12, 1.93])
        self.label_predictions = torch.Tensor([0.0, 0.0, 1.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0, 1.0])
        self.targets = torch.Tensor([1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 1.0, 1.0, 0.0])
        self.binary_accuracy_none = (self.label_predictions == self.targets).float() * 100
        self.binary_accuracy = self.binary_accuracy_none.mean()

    def test_standard(self):
        binary_accuracy = BinaryAccuracy()
        actual = binary_accuracy(self.predictions, self.targets)
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal(self.binary_accuracy, actual, decimal=5)

    def test_functional(self):
        actual = bin_acc(self.predictions, self.targets)
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal(self.binary_accuracy, actual, decimal=5)

    def test_hundred(self):
        binary_accuracy = BinaryAccuracy()
        actual = binary_accuracy(self.predictions, self.label_predictions)
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal(100.0, actual, decimal=5)

    def test_sum(self):
        binary_accuracy = BinaryAccuracy(reduction='sum')
        actual = binary_accuracy(self.predictions, self.targets)
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal(self.binary_accuracy_none.sum(), actual)

    def test_functional_sum(self):
        actual = bin_acc(self.predictions, self.targets, reduction='sum')
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal(self.binary_accuracy_none.sum(), actual)

    def test_sum_hundred(self):
        binary_accuracy = BinaryAccuracy(reduction='sum')
        actual = binary_accuracy(self.predictions, self.label_predictions)
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal(100.0 * len(self.predictions), actual)

    def test_none(self):
        binary_accuracy = BinaryAccuracy(reduction='none')
        actual = binary_accuracy(self.predictions, self.targets)
        self.assertEqual(self.targets.shape, actual.shape)
        self.assertTrue(torch.all(self.binary_accuracy_none == actual))

    def test_functional_none(self):
        actual = bin_acc(self.predictions, self.targets, reduction='none')
        self.assertEqual(self.targets.shape, actual.shape)
        self.assertTrue(torch.all(self.binary_accuracy_none == actual))

    def test_none_hundred(self):
        binary_accuracy = BinaryAccuracy(reduction='none')
        actual = binary_accuracy(self.predictions, self.label_predictions)
        self.assertEqual(self.targets.shape, actual.shape)
        self.assertTrue(torch.all(100.0 * torch.ones_like(self.label_predictions).float() == actual))

    def test_threshold(self):
        threshold = -0.5
        binary_accuracy = BinaryAccuracy(threshold=threshold)
        actual = binary_accuracy(self.predictions, self.targets)
        label_predictions = (self.predictions > threshold).float()
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal((label_predictions == self.targets).float().mean() * 100, actual)

    def test_functional_threshold(self):
        threshold = -0.5
        actual = bin_acc(self.predictions, self.targets, threshold=threshold)
        label_predictions = (self.predictions > threshold).float()
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal((label_predictions == self.targets).float().mean() * 100, actual)

    @skipIf(not torch.cuda.is_available(), "no gpu available")
    def test_on_gpu(self):
        binary_accuracy = BinaryAccuracy()
        binary_accuracy.cuda()
        actual = binary_accuracy(self.predictions.cuda(), self.targets.cuda())
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal(self.binary_accuracy, actual.cpu(), decimal=5)

    @skipIf(not torch.cuda.is_available(), "no gpu available")
    def test_functional_on_gpu(self):
        actual = bin_acc(self.predictions.cuda(), self.targets.cuda())
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal(self.binary_accuracy, actual.cpu(), decimal=5)


class TopKAccuracyTest(TestCase):
    def setUp(self):
        self.predictions = torch.Tensor(
            [
                [0.35, 0.25, 0.1, 0.1, 0.2],
                [0.1, 0.7, 0.0, 0.2, 0.0],
                [0.0, 0.7, 0.1, 0.2, 0.0],
                [0.1, 0.5, 0.0, 0.3, 0.0],
                [0.0, 0.3, 0.1, 0.7, 0.0],
                [0.1, 0.6, 0.0, 0.3, 0.0],
            ]
        )
        self.top_3_label_predictions = torch.Tensor([[0, 1, 4], [1, 3, 0], [1, 3, 0], [1, 3, 0], [3, 1, 2], [1, 3, 0]])
        self.targets = torch.Tensor([0, 4, 1, 2, 3, 0])
        self.top_k_acc_none = (self.top_3_label_predictions == self.targets.unsqueeze(1)).any(1).float() * 100
        self.top_k_acc = self.top_k_acc_none.mean()

    def test_standard(self):
        top_k_acc = TopKAccuracy(3)
        actual = top_k_acc(self.predictions, self.targets)
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal(self.top_k_acc, actual, decimal=5)

    def test_functional(self):
        actual = topk(self.predictions, self.targets, 3)
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal(self.top_k_acc, actual, decimal=5)

    def test_hundred(self):
        top_k_acc = TopKAccuracy(3)
        actual = top_k_acc(self.predictions, self.top_3_label_predictions[:, 0])
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal(100.0, actual)

    def test_sum(self):
        top_k_acc = TopKAccuracy(3, reduction='sum')
        actual = top_k_acc(self.predictions, self.targets)
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal(self.top_k_acc * len(self.predictions), actual, decimal=0)

    def test_functional_sum(self):
        actual = topk(self.predictions, self.targets, 3, reduction='sum')
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal(self.top_k_acc * len(self.predictions), actual, decimal=0)

    def test_sum_hundred(self):
        top_k_acc = TopKAccuracy(3, reduction='sum')
        actual = top_k_acc(self.predictions, self.top_3_label_predictions[:, 0])
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal(100.0 * len(self.predictions), actual, decimal=0)

    def test_none(self):
        top_k_acc = TopKAccuracy(3, reduction='none')
        actual = top_k_acc(self.predictions, self.targets)
        self.assertEqual(self.targets.shape, actual.shape)
        self.assertTrue(torch.all(self.top_k_acc_none == actual))

    def test_functional_none(self):
        actual = topk(self.predictions, self.targets, 3, reduction='none')
        self.assertEqual(self.targets.shape, actual.shape)
        self.assertTrue(torch.all(self.top_k_acc_none == actual))

    def test_none_hundred(self):
        top_k_acc = TopKAccuracy(3, reduction='none')
        actual = top_k_acc(self.predictions, self.top_3_label_predictions[:, 0])
        self.assertEqual(self.targets.shape, actual.shape)
        self.assertTrue(torch.all(100.0 * torch.ones_like(self.targets).float() == actual))

    @skipIf(not torch.cuda.is_available(), "no gpu available")
    def test_on_gpu(self):
        top_k_acc = TopKAccuracy(3)
        top_k_acc.cuda()
        actual = top_k_acc(self.predictions.cuda(), self.targets.cuda())
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal(self.top_k_acc, actual.cpu(), decimal=5)

    @skipIf(not torch.cuda.is_available(), "no gpu available")
    def test_functional_on_gpu(self):
        actual = topk(self.predictions.cuda(), self.targets.cuda(), 3)
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal(self.top_k_acc, actual.cpu(), decimal=5)


class IgnoreIndexTopKAccuracyTest(TestCase):
    def setUp(self):
        self.predictions = torch.Tensor(
            [
                [0.35, 0.25, 0.1, 0.1, 0.2],
                [0.1, 0.7, 0.0, 0.2, 0.0],
                [0.0, 0.7, 0.1, 0.2, 0.0],
                [0.1, 0.5, 0.0, 0.3, 0.0],
                [0.0, 0.3, 0.1, 0.7, 0.0],
                [0.1, 0.6, 0.0, 0.3, 0.0],
            ]
        )
        self.top_3_label_predictions = torch.Tensor([[0, 1, 4], [1, 3, 0], [1, 3, 0], [1, 3, 0], [3, 1, 2], [1, 3, 0]])
        self.targets = torch.Tensor([-100, 4, 1, -100, 3, 0])
        self.top_k_acc = 75.0

    def test_standard(self):
        top_k_acc = TopKAccuracy(3)
        actual = top_k_acc(self.predictions, self.targets)
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal(self.top_k_acc, actual)

    def test_functional_standard(self):
        actual = topk(self.predictions, self.targets, 3)
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal(self.top_k_acc, actual)

    def test_ignore_index_with_different_value(self):
        top_k_acc = TopKAccuracy(3, ignore_index=-1)
        actual = top_k_acc(self.predictions, self.targets)
        self.assertEqual((), actual.shape)
        self.assertAlmostEqual(50.0, float(actual), places=5)

    def test_functional_ignore_index_with_different_value(self):
        actual = topk(self.predictions, self.targets, 3, ignore_index=-1)
        self.assertEqual((), actual.shape)
        self.assertAlmostEqual(50.0, float(actual), places=5)

    def test_ignore_index_with_valid_index(self):
        top_k_acc = TopKAccuracy(3, ignore_index=1)
        actual = top_k_acc(self.predictions, self.targets)
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal(40.0, actual)

    def test_functional_ignore_index_with_valid_index(self):
        actual = topk(self.predictions, self.targets, 3, ignore_index=1)
        self.assertEqual((), actual.shape)
        np.testing.assert_almost_equal(40.0, actual)



import os
from io import BytesIO
from tempfile import TemporaryDirectory
from unittest import TestCase
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.axes import Axes
from PIL import Image

from poutyne import plot_history, plot_metric


class PlotHistoryTest(TestCase):
    HISTORY = [
        {'epoch': 1, 'time': 6.2788, 'loss': 0.3683, 'acc': 88.2645, 'val_loss': 0.0984, 'val_acc': 97.0833},
        {'epoch': 2, 'time': 6.2570, 'loss': 0.1365, 'acc': 95.9166, 'val_loss': 0.0680, 'val_acc': 97.9916},
        {'epoch': 3, 'time': 6.3314, 'loss': 0.1057, 'acc': 96.8458, 'val_loss': 0.0531, 'val_acc': 98.4916},
        {'epoch': 4, 'time': 6.2574, 'loss': 0.0874, 'acc': 97.2937, 'val_loss': 0.0521, 'val_acc': 98.4166},
        {'epoch': 5, 'time': 6.3268, 'loss': 0.0775, 'acc': 97.6666, 'val_loss': 0.0464, 'val_acc': 98.6166},
    ]
    METRICS = ['time', 'loss', 'acc']
    NUM_METRIC_PLOTS = len(METRICS)

    def test_basic(self):
        figs, axes = plot_history(PlotHistoryTest.HISTORY)
        self.assertEqual(len(figs), PlotHistoryTest.NUM_METRIC_PLOTS)
        self.assertEqual(len(axes), PlotHistoryTest.NUM_METRIC_PLOTS)
        for fig, ax in zip(figs, axes):
            self.assertIsInstance(fig, Figure)
            self.assertIsInstance(ax, Axes)

    def test_compare_plot_history_plot_metric(self):
        plot_history_figs, _ = plot_history(PlotHistoryTest.HISTORY, close=False, show=False)

        plot_metrics_figs = []
        for metric in PlotHistoryTest.METRICS:
            fig, ax = plt.subplots()
            plot_metric(PlotHistoryTest.HISTORY, metric, ax=ax)
            plot_metrics_figs.append(fig)

        for plot_history_fig, plot_metrics_fig in zip(plot_history_figs, plot_metrics_figs):
            self.assertEqual(self._to_image(plot_history_fig), self._to_image(plot_metrics_fig))

    def test_all_different(self):
        plot_history_figs, _ = plot_history(PlotHistoryTest.HISTORY, close=False, show=False)
        images = list(map(self._to_image, plot_history_figs))
        for i, _ in enumerate(images):
            for j in range(i + 1, len(images)):
                self.assertNotEqual(images[i], images[j])

    def test_title(self):
        title = 'My Title'
        plot_history_figs, _ = plot_history(PlotHistoryTest.HISTORY, titles=title, close=False, show=False)

        plot_metrics_figs_no_title = []
        for metric in PlotHistoryTest.METRICS:
            fig, ax = plt.subplots()
            plot_metric(PlotHistoryTest.HISTORY, metric, ax=ax)
            plot_metrics_figs_no_title.append(fig)

        plot_metrics_figs_with_title = []
        for metric in PlotHistoryTest.METRICS:
            fig, ax = plt.subplots()
            plot_metric(PlotHistoryTest.HISTORY, metric, title=title, ax=ax)
            plot_metrics_figs_with_title.append(fig)

        for (
            plot_history_fig,
            plot_metrics_fig_no_title,
            plot_metrics_fig_with_title,
        ) in zip(plot_history_figs, plot_metrics_figs_no_title, plot_metrics_figs_with_title):
            self.assertEqual(self._to_image(plot_history_fig), self._to_image(plot_metrics_fig_with_title))
            self.assertNotEqual(self._to_image(plot_history_fig), self._to_image(plot_metrics_fig_no_title))

    def test_different_titles(self):
        titles = ['Time', 'Loss', 'Accuracy']
        plot_history_figs, _ = plot_history(PlotHistoryTest.HISTORY, titles=titles, close=False, show=False)

        plot_metrics_figs_no_title = []
        for metric in PlotHistoryTest.METRICS:
            fig, ax = plt.subplots()
            plot_metric(PlotHistoryTest.HISTORY, metric, ax=ax)
            plot_metrics_figs_no_title.append(fig)

        plot_metrics_figs_with_title = []
        for metric, title in zip(PlotHistoryTest.METRICS, titles):
            fig, ax = plt.subplots()
            plot_metric(PlotHistoryTest.HISTORY, metric, title=title, ax=ax)
            plot_metrics_figs_with_title.append(fig)

        for (
            plot_history_fig,
            plot_metrics_fig_no_title,
            plot_metrics_fig_with_title,
        ) in zip(plot_history_figs, plot_metrics_figs_no_title, plot_metrics_figs_with_title):
            self.assertEqual(self._to_image(plot_history_fig), self._to_image(plot_metrics_fig_with_title))
            self.assertNotEqual(self._to_image(plot_history_fig), self._to_image(plot_metrics_fig_no_title))

    def test_labels(self):
        labels = ['Time', 'Loss', 'Accuracy']
        plot_history_figs, _ = plot_history(PlotHistoryTest.HISTORY, labels=labels, close=False, show=False)

        plot_metrics_figs_no_labels = []
        for metric in PlotHistoryTest.METRICS:
            fig, ax = plt.subplots()
            plot_metric(PlotHistoryTest.HISTORY, metric, ax=ax)
            plot_metrics_figs_no_labels.append(fig)

        plot_metrics_figs_with_labels = []
        for metric, label in zip(PlotHistoryTest.METRICS, labels):
            fig, ax = plt.subplots()
            plot_metric(PlotHistoryTest.HISTORY, metric, label=label, ax=ax)
            plot_metrics_figs_with_labels.append(fig)

        for (
            plot_history_fig,
            plot_metrics_fig_no_labels,
            plot_metrics_fig_with_labels,
        ) in zip(plot_history_figs, plot_metrics_figs_no_labels, plot_metrics_figs_with_labels):
            self.assertEqual(self._to_image(plot_history_fig), self._to_image(plot_metrics_fig_with_labels))
            self.assertNotEqual(self._to_image(plot_history_fig), self._to_image(plot_metrics_fig_no_labels))

    def test_plot_metrics_use_gca(self):
        fig, ax = plt.subplots()
        plot_metric(PlotHistoryTest.HISTORY, 'loss', ax=ax)
        image_with_provided_ax = self._to_image(fig)

        plot_metric(PlotHistoryTest.HISTORY, 'loss')
        image_with_gca = self._to_image(plt.gcf())

        self.assertEqual(image_with_provided_ax, image_with_gca)

    def test_with_provided_axes(self):
        provided_figs, provided_axes = zip(*(plt.subplots() for _ in range(PlotHistoryTest.NUM_METRIC_PLOTS)))
        ret_figs, ret_axes = plot_history(PlotHistoryTest.HISTORY, close=False, show=False, axes=provided_axes)

        new_figs, _ = plot_history(PlotHistoryTest.HISTORY, close=False, show=False)

        self.assertEqual(ret_axes, provided_axes)
        self.assertEqual(len(ret_figs), 0)

        for provided_fig, new_fig in zip(provided_figs, new_figs):
            self.assertEqual(self._to_image(provided_fig), self._to_image(new_fig))

    def test_save(self):
        temp_dir_obj = TemporaryDirectory()
        path = temp_dir_obj.name
        filename = 'test_{metric}'
        final_png_filenames = [
            os.path.join(path, f'{filename.format(metric=metric)}.png') for metric in PlotHistoryTest.METRICS
        ]
        final_pdf_filenames = [
            os.path.join(path, f'{filename.format(metric=metric)}.pdf') for metric in PlotHistoryTest.METRICS
        ]

        figs, _ = plot_history(
            PlotHistoryTest.HISTORY,
            close=False,
            show=False,
            save=True,
            save_directory=path,
            save_filename_template=filename,
            save_extensions=('png', 'pdf'),
        )

        for png_filename, pdf_filename in zip(final_png_filenames, final_pdf_filenames):
            self.assertTrue(os.path.isfile(png_filename))
            self.assertTrue(os.path.isfile(pdf_filename))

        saved_images = [Image.open(filename) for filename in final_png_filenames]
        ret_images = [self._to_image(fig) for fig in figs]
        for save_image, ret_image in zip(saved_images, ret_images):
            self.assertEqual(save_image, ret_image)

    def _to_image(self, fig):
        fd = BytesIO()
        fig.savefig(fd, format='png')
        image = Image.open(fd)
        plt.close(fig)
        return image



import unittest

from poutyne import linspace, cosinespace, Phase, OptimizerPolicy, one_cycle_phases, sgdr_phases


class TestSpaces(unittest.TestCase):
    def assert_space(self, space, expected):
        for val, exp in zip(space, expected):
            self.assertAlmostEqual(val, exp, places=3)

    def test_linspace_const(self):
        self.assert_space(linspace(0, 0, 3), [0, 0, 0])

    def test_linspace_increasing(self):
        self.assert_space(linspace(0, 1, 3), [0, 0.5, 1])

    def test_linspace_decreasing(self):
        self.assert_space(linspace(1, 0, 3), [1, 0.5, 0])

    def test_linspace_with_many_values(self):
        self.assert_space(linspace(0, 1, 6), [0, 0.2, 0.4, 0.6, 0.8, 1])

    def test_cosinespace_const(self):
        self.assert_space(cosinespace(0, 0, 2), [0, 0])

    def test_cosinespace_increasing(self):
        self.assert_space(cosinespace(0, 1, 2), [0, 1])
        self.assert_space(cosinespace(0, 1, 3), [0, 0.5, 1])

    def test_cosinespace_decreasing(self):
        self.assert_space(cosinespace(1, 0, 2), [1, 0])
        self.assert_space(cosinespace(1, 0, 3), [1, 0.5, 0])

    def test_cosinespace_with_many_values(self):
        self.assert_space(cosinespace(0, 1, 5), [0, 0.1464, 0.5, 0.8535, 1])
        self.assert_space(cosinespace(1, 0, 5), [1, 0.8535, 0.5, 0.1464, 0])

    def test_space_has_desired_legth(self):
        for space_fn in [linspace, cosinespace]:
            space = list(space_fn(2, 1, 3))
            assert len(space) == 3
            with self.assertRaises(IndexError):
                space[4]  # pylint: disable=pointless-statement


class TestPhase(unittest.TestCase):
    def test_init_raises_without_lr_or_momentum(self):
        with self.assertRaises(ValueError):
            Phase(lr=None, momentum=None)
        with self.assertRaises(ValueError):
            Phase()

    def test_phase_with_only_one_parameter_set(self):
        for param_name in ["lr", "momentum"]:
            steps = 3
            phase = Phase(**{param_name: linspace(1, 0, steps)})
            for params in phase:
                self.assertIsInstance(params, dict)
                self.assertTrue(param_name in params)
                self.assertEqual(len(params), 1)
                self.assertTrue(0 <= params[param_name] <= 1)

    def test_phase_with_two_parameters(self):
        steps = 4
        phase = Phase(lr=linspace(1, 0, steps), momentum=cosinespace(0.8, 1, steps))
        self.assertEqual(len(list(phase)), steps)
        for params in phase:
            self.assertEqual(len(params), 2)

            self.assertTrue("lr" in params)
            self.assertTrue(0 <= params["lr"] <= 1)

            self.assertTrue("momentum" in params)
            self.assertTrue(0.8 <= params["momentum"] <= 1)


class TestOptimizerPolicy(unittest.TestCase):
    def setUp(self):
        steps = 2
        phases = [Phase(lr=linspace(1, 1, steps)), Phase(lr=linspace(0, 0, steps))]
        self.policy = OptimizerPolicy(phases)

    def test_basic_iteration(self):
        policy_iter = iter(self.policy)
        self.assertEqual(next(policy_iter), {"lr": 1})
        self.assertEqual(next(policy_iter), {"lr": 1})

        self.assertEqual(next(policy_iter), {"lr": 0})
        self.assertEqual(next(policy_iter), {"lr": 0})

        with self.assertRaises(StopIteration):
            next(policy_iter)


class TestOptimizerPolicyRestart(unittest.TestCase):
    def setUp(self):
        steps = 2
        phases = [
            Phase(lr=linspace(0, 0, steps)),
            Phase(lr=linspace(1, 1, steps)),
            Phase(lr=linspace(2, 2, steps)),
        ]
        self.policy = OptimizerPolicy(phases=phases, initial_step=3)

    def test_starts_at_correct_position(self):
        policy_iter = iter(self.policy)
        # The first three steps are ignored
        # assert next(policy_iter) == {"lr": 0}
        # assert next(policy_iter) == {"lr": 0}
        # assert next(policy_iter) == {"lr": 1}
        assert next(policy_iter) == {"lr": 1}
        assert next(policy_iter) == {"lr": 2}
        assert next(policy_iter) == {"lr": 2}

        with self.assertRaises(StopIteration):
            next(policy_iter)


class TestOneCycle(unittest.TestCase):
    def test_length(self):
        policy = OptimizerPolicy(one_cycle_phases(100))
        self.assertEqual(len(list(policy.all_steps())), 100)

        policy = OptimizerPolicy(one_cycle_phases(99))
        self.assertEqual(len(list(policy.all_steps())), 99)


class TestSGDR(unittest.TestCase):
    def test_length_with_cycle_mult_one(self):
        policy = OptimizerPolicy(sgdr_phases(10, 1, cycle_mult=1))
        self.assertEqual(len(list(policy.all_steps())), 10)

        policy = OptimizerPolicy(sgdr_phases(10, 2, cycle_mult=1))
        self.assertEqual(len(list(policy.all_steps())), 20)

        policy = OptimizerPolicy(sgdr_phases(10, 10, cycle_mult=1))
        self.assertEqual(len(list(policy.all_steps())), 100)

    def test_length_with_higher_cycle_mult(self):
        policy = OptimizerPolicy(sgdr_phases(10, 1, cycle_mult=2))
        self.assertEqual(len(list(policy.all_steps())), 10)

        policy = OptimizerPolicy(sgdr_phases(10, 2, cycle_mult=2))
        self.assertEqual(len(list(policy.all_steps())), 30)

        policy = OptimizerPolicy(sgdr_phases(10, 3, cycle_mult=2))
        self.assertEqual(len(list(policy.all_steps())), 70)

        policy = OptimizerPolicy(sgdr_phases(10, 1, cycle_mult=3))
        self.assertEqual(len(list(policy.all_steps())), 10)

        policy = OptimizerPolicy(sgdr_phases(10, 2, cycle_mult=3))
        self.assertEqual(len(list(policy.all_steps())), 40)



import os

import unittest
from unittest import TestCase

from tempfile import TemporaryDirectory

import torch
import torch.nn as nn

from poutyne import torch_to_numpy, Model, ModelCheckpoint
from tests.framework.tools import some_data_generator


class ModelCheckpointTest(TestCase):
    batch_size = 20

    def setUp(self):
        torch.manual_seed(42)
        self.pytorch_network = nn.Linear(1, 1)
        self.loss_function = nn.MSELoss()
        self.optimizer = torch.optim.SGD(self.pytorch_network.parameters(), lr=1e-3)
        self.model = Model(self.pytorch_network, self.optimizer, self.loss_function)
        self.temp_dir_obj = TemporaryDirectory()
        self.checkpoint_filename = os.path.join(self.temp_dir_obj.name, 'my_checkpoint_{epoch}.ckpt')

    def tearDown(self):
        self.temp_dir_obj.cleanup()

    def test_integration(self):
        train_gen = some_data_generator(ModelCheckpointTest.batch_size)
        valid_gen = some_data_generator(ModelCheckpointTest.batch_size)
        checkpointer = ModelCheckpoint(self.checkpoint_filename, monitor='val_loss', verbose=True, save_best_only=True)
        self.model.fit_generator(train_gen, valid_gen, epochs=10, steps_per_epoch=5, callbacks=[checkpointer])

    def test_temporary_filename_arg(self):
        tmp_filename = os.path.join(self.temp_dir_obj.name, 'my_checkpoint.tmp.ckpt')
        checkpoint_filename = os.path.join(self.temp_dir_obj.name, 'my_checkpoint.ckpt')
        train_gen = some_data_generator(ModelCheckpointTest.batch_size)
        valid_gen = some_data_generator(ModelCheckpointTest.batch_size)
        checkpointer = ModelCheckpoint(
            checkpoint_filename, monitor='val_loss', verbose=True, period=1, temporary_filename=tmp_filename
        )
        self.model.fit_generator(train_gen, valid_gen, epochs=10, steps_per_epoch=5, callbacks=[checkpointer])
        self.assertFalse(os.path.isfile(tmp_filename))
        self.assertTrue(os.path.isfile(checkpoint_filename))

    def test_temporary_filename_arg_with_differing_checkpoint_filename(self):
        epochs = 10
        tmp_filename = os.path.join(self.temp_dir_obj.name, 'my_checkpoint.tmp.ckpt')
        checkpoint_filename = os.path.join(self.temp_dir_obj.name, 'my_checkpoint_{epoch}.ckpt')
        train_gen = some_data_generator(ModelCheckpointTest.batch_size)
        valid_gen = some_data_generator(ModelCheckpointTest.batch_size)
        checkpointer = ModelCheckpoint(
            checkpoint_filename, monitor='val_loss', verbose=True, period=1, temporary_filename=tmp_filename
        )
        self.model.fit_generator(train_gen, valid_gen, epochs=epochs, steps_per_epoch=5, callbacks=[checkpointer])
        self.assertFalse(os.path.isfile(tmp_filename))
        for i in range(1, epochs + 1):
            self.assertTrue(os.path.isfile(checkpoint_filename.format(epoch=i)))

    def test_non_atomic_write(self):
        checkpoint_filename = os.path.join(self.temp_dir_obj.name, 'my_checkpoint.ckpt')
        train_gen = some_data_generator(ModelCheckpointTest.batch_size)
        valid_gen = some_data_generator(ModelCheckpointTest.batch_size)
        checkpointer = ModelCheckpoint(
            checkpoint_filename, monitor='val_loss', verbose=True, period=1, atomic_write=False
        )
        self.model.fit_generator(train_gen, valid_gen, epochs=10, steps_per_epoch=5, callbacks=[checkpointer])
        self.assertTrue(os.path.isfile(checkpoint_filename))

    def test_save_best_only(self):
        checkpointer = ModelCheckpoint(self.checkpoint_filename, monitor='val_loss', verbose=True, save_best_only=True)

        val_losses = [10, 3, 8, 5, 2]
        has_checkpoints = [True, True, False, False, True]
        self._test_checkpointer_with_val_losses(checkpointer, val_losses, has_checkpoints)

    def test_save_best_only_with_restore_best(self):
        checkpointer = ModelCheckpoint(
            self.checkpoint_filename, monitor='val_loss', verbose=True, save_best_only=True, restore_best=True
        )

        val_losses = [10, 3, 8, 5, 2]
        has_checkpoints = [True, True, False, False, True]
        self._test_checkpointer_with_val_losses(checkpointer, val_losses, has_checkpoints)

        self._test_restore_best(val_losses)

    def test_restore_best_without_save_best_only(self):
        with self.assertRaises(ValueError):
            ModelCheckpoint(
                self.checkpoint_filename, monitor='val_loss', verbose=True, save_best_only=False, restore_best=True
            )

        with self.assertRaises(ValueError):
            ModelCheckpoint(self.checkpoint_filename, monitor='val_loss', verbose=True, restore_best=True)

    def test_save_best_only_with_max(self):
        checkpointer = ModelCheckpoint(
            self.checkpoint_filename, monitor='val_loss', mode='max', verbose=True, save_best_only=True
        )

        val_losses = [2, 3, 8, 5, 2]
        has_checkpoints = [True, True, True, False, False]
        self._test_checkpointer_with_val_losses(checkpointer, val_losses, has_checkpoints)

    def test_periodic_with_period_of_1(self):
        checkpointer = ModelCheckpoint(
            self.checkpoint_filename, monitor='val_loss', verbose=True, period=1, save_best_only=False
        )

        val_losses = [1] * 10
        has_checkpoints = [True] * 10
        self._test_checkpointer_with_val_losses(checkpointer, val_losses, has_checkpoints)

    def test_periodic_with_period_of_2(self):
        checkpointer = ModelCheckpoint(
            self.checkpoint_filename, monitor='val_loss', verbose=True, period=2, save_best_only=False
        )

        val_losses = [1] * 10
        has_checkpoints = [False, True] * 5
        self._test_checkpointer_with_val_losses(checkpointer, val_losses, has_checkpoints)

    def _test_checkpointer_with_val_losses(self, checkpointer, val_losses, has_checkpoints):
        generator = some_data_generator(ModelCheckpointTest.batch_size)

        checkpointer.set_params({'epochs': len(val_losses), 'steps': 1})
        checkpointer.set_model(self.model)
        checkpointer.on_train_begin({})
        for epoch, (val_loss, has_checkpoint) in enumerate(zip(val_losses, has_checkpoints), 1):
            checkpointer.on_epoch_begin(epoch, {})
            checkpointer.on_train_batch_begin(1, {})
            loss = self._update_model(generator)
            checkpointer.on_train_batch_end(1, {'batch': 1, 'size': ModelCheckpointTest.batch_size, 'loss': loss})
            checkpointer.on_epoch_end(epoch, {'epoch': epoch, 'loss': loss, 'val_loss': val_loss})
            filename = self.checkpoint_filename.format(epoch=epoch)
            self.assertEqual(has_checkpoint, os.path.isfile(filename))
        checkpointer.on_train_end({})

    def _update_model(self, generator):
        self.pytorch_network.zero_grad()

        x, y = next(generator)
        pred_y = self.pytorch_network(x)
        loss = self.loss_function(pred_y, y)
        loss.backward()

        self.optimizer.step()

        return float(loss)

    def _test_restore_best(self, val_losses):
        final_weights = torch_to_numpy(self.model.get_weight_copies())

        epoch = val_losses.index(min(val_losses)) + 1
        best_epoch_filename = self.checkpoint_filename.format(epoch=epoch)
        self.model.load_weights(best_epoch_filename)

        best_weights = torch_to_numpy(self.model.get_weight_copies())

        self.assertEqual(best_weights, final_weights)


if __name__ == '__main__':
    unittest.main()



# pylint: disable=line-too-long, pointless-string-statement
import os
import warnings
from typing import Dict, Union, Mapping, Sequence

from . import Logger

try:
    import mlflow
    from mlflow.exceptions import MlflowException
    from mlflow.tracking import MlflowClient
except ImportError:
    mlflow = None

try:
    import git
except ImportError:
    git = None


class MLFlowLogger(Logger):
    """
    MLflow logger to manage logging of experiments parameters, metrics update, models log and other information. The
    logger will log all run into the same experiment.

    Args:
        experiment_name (str): The name of the experiment. Name must be unique and are case sensitive.
        tracking_uri (Union[str, None]): Either the URI tracking path (for server tracking) of the absolute path to
            the directory to save the files (for file store). For example: ``http://<ip address>:<port>``
            (remote server) or ``/home/<user>/mlflow-server`` (local server).
            If None, will use the default MLflow file tracking URI ``"./mlruns"``.
        batch_granularity (bool): Whether to also output the result of each batch in addition to the epochs.
            (Default value = False)

    Example:
        Using file store::

            mlflow_logger = MLFlowLogger(experiment_name="experiment", tracking_uri="/absolute/path/to/directory")
            mlflow_logger.log_config_params(config_params=cfg_dict) # logging the config dictionary

            # our Poutyne experiment
            experiment = Experiment(directory=saving_directory, network=network, device=device, optimizer=optimizer,
                            loss_function=cross_entropy_loss, batch_metrics=[accuracy])

            # Using the MLflow logger callback during training
            experiment.train(train_generator=train_loader, valid_generator=valid_loader, epochs=1,
                             seed=42, callbacks=[mlflow_logger])

        Using server tracking::

            mlflow_logger = MLFlowLogger(experiment_name="experiment", tracking_uri="http://IP_ADDRESS:PORT")
            mlflow_logger.log_config_params(config_params=cfg_dict) # logging the config dictionary

            # our Poutyne experiment
            experiment = Experiment(directory=saving_directory, network=network, device=device, optimizer=optimizer,
                            loss_function=cross_entropy_loss, batch_metrics=[accuracy])

            # Using the MLflow logger callback during training
            experiment.train(train_generator=train_loader, valid_generator=valid_loader, epochs=1,
                             seed=42, callbacks=[mlflow_logger])
    """

    def __init__(
        self, experiment_name: str, tracking_uri: Union[str, None] = None, batch_granularity: bool = False
    ) -> None:
        super().__init__(batch_granularity=batch_granularity)
        if mlflow is None:
            raise ImportError("Mlflow needs to be installed to use this callback.")

        self.tracking = tracking_uri

        self._working_directory = os.getcwd()  # for Git

        self.ml_flow_client = MlflowClient(tracking_uri=self.tracking)

        self._handle_experiment_id(experiment_name)
        self.run_id = self.ml_flow_client.create_run(experiment_id=self.experiment_id).info.run_id

        self._log_git_version()

        self._status = "FAILED"  # base case is a failure

    def log_config_params(self, config_params: Union[Mapping, Sequence]) -> None:
        """
        Args:
            config_params (Union[Mapping, Sequence]):
                The config parameters of the training to log, such as number of epoch, loss function, optimizer etc.
        """
        if isinstance(config_params, Mapping):
            for param_name, element in config_params.items():
                self._log_config_write(param_name, element)
        else:  # equivalent to "if isinstance(config_params, Sequence):"
            for idx, element in enumerate(config_params):
                self._log_config_write(str(idx), element)

    def log_param(self, param_name: str, value: Union[str, float]) -> None:
        """
        Log the value of a parameter into the experiment.

        Args:
            param_name (str): The name of the parameter.
            value (Union[str, float]: The value of the parameter.

        """
        self.ml_flow_client.log_param(run_id=self.run_id, key=param_name, value=value)

    def log_metric(self, metric_name: str, value: float, step: Union[int, None] = None) -> None:
        """
        Log the value of a metric into the experiment.

        Args:
            metric_name (str): The name of the metric.
            value (float): The value of the metric.
            step (Union[int, None]): The step when the metric was compute (Default = None).
        """
        self.ml_flow_client.log_metric(run_id=self.run_id, key=metric_name, value=value, step=step)

    def _log_config_write(self, parent_name: str, element: Union[int, float, str, Mapping, Sequence]) -> None:
        """
        Log the config parameters when it's a mapping or a sequence of elements.
        """
        if isinstance(element, Mapping):
            for key, value in element.items():
                # We recursively open the element (Dict format type)
                self._log_config_write(f"{parent_name}.{key}", value)
        elif isinstance(element, Sequence) and not isinstance(element, str):
            # Since str are sequence we negate it to be log in the else
            for idx, value in enumerate(element):
                self._log_config_write(f"{parent_name}.{idx}", value)
        else:
            self.log_param(parent_name, element)

    def _on_train_batch_end_write(self, batch_number: int, logs: Dict) -> None:
        """
        Log the batch metric.
        """
        for key, value in logs.items():
            self.log_metric(key, value, step=batch_number)

    def _on_epoch_end_write(self, epoch_number: int, logs: Dict) -> None:
        """
        Log the batch and epoch metric.
        """
        logs.pop("epoch")
        for key, value in logs.items():
            self.log_metric(key, value, step=epoch_number)

    def on_train_end(self, logs: Dict):
        """
        Log the last epoch batch and epoch metric and close the active run.
        """
        self._on_train_end_write(logs)
        self._status = "FINISHED"

        mlflow.end_run()
        self._status_handling()

    def _on_train_end_write(self, logs) -> None:
        """
        Log the last epoch batch and epoch metric.
        """
        last_epoch = self.params["epochs"]
        self.log_metric("last-epoch", last_epoch)

    def on_test_begin(self, logs: Dict):
        self._status = "FAILED"  # to change status from FINISHED to FAILED (base case) if trained before
        self._status_handling()

    def on_test_end(self, logs: Dict):
        """
        Log the test results.
        """
        if len(logs) > 0:  # to manage failure of the test loop
            self._on_test_end_write(logs)
            self._status = "FINISHED"

        mlflow.end_run()
        self._status_handling()

    def _on_test_end_write(self, logs: Dict) -> None:
        for key, value in logs.items():
            self.log_metric(key, value)

    def _status_handling(self):
        # we set_terminated the run to get the finishing status (FINISHED or FAILED)
        self.ml_flow_client.set_terminated(self.run_id, status=self._status)

    def _handle_experiment_id(self, experiment_name):
        """
        Handle if the experiment name already exist to grad the id and append new experiment to it.
        """
        try:
            self.experiment_id = self.ml_flow_client.create_experiment(experiment_name, self.tracking)
        except MlflowException:
            self.experiment_id = self.ml_flow_client.get_experiment_by_name(experiment_name).experiment_id

    def _log_git_version(self):
        """
        Log the git commit of the run.
        """
        source_version = _get_git_commit(self._working_directory)
        if source_version is not None:
            self.ml_flow_client.set_tag(self.run_id, "mlflow.source.git.commit", source_version)


"""
The source code of the function _get_git_commit was inspired from the MLflow project
(https://github.com/mlflow/mlflow/blob/7fde53e497c50b4eb4da1e9988710695b8c2e093/mlflow/tracking/context/git_context.py#L11),
and has been modified. All modifications made from the original source code are under the LGPLv3 license.

COPYRIGHT

All contributions by MLflow:
Copyright (c) 2020, MLFlow.
All rights reserved.

Each contributor holds copyright over their respective contributions. The project versioning (Git)
records all such contribution source information.
"""


def _get_git_commit(path):
    """
    Function to get the git commit from a path.
    """
    if git is None:
        warnings.warn(
            "Failed to import Git (the Git executable is probably not on your PATH)," " so Git SHA is not available."
        )
        return None

    try:
        repo = git.Repo(path, search_parent_directories=True)
        commit = repo.head.commit.hexsha
        return commit
    except (git.InvalidGitRepositoryError, git.NoSuchPathError) as e:
        warnings.warn(f"Failed to grab the git repository so Git SHA is not available. Error: {e}")
        return None



# -*- coding: utf-8 -*-
import random
import numpy as np
import torch
from torch.utils.data import Dataset


def torch_to_numpy(obj, copy=False):
    """
    Convert to Numpy arrays all tensors inside a Python object composed of the supported types.

    Args:
        obj: The Python object to convert.
        copy (bool): Whether to copy the memory. By default, if a tensor is already on CPU, the
            Numpy array will be a view of the tensor.

    Returns:
        A new Python object with the same structure as `obj` but where the tensors are now Numpy
        arrays. Not supported type are left as reference in the new object.

    Example:
        .. code-block:: python

            >>> from poutyne import torch_to_numpy
            >>> torch_to_numpy({
            ...     'first': torch.tensor([1, 2, 3]),
            ...     'second':[torch.tensor([4,5,6]), torch.tensor([7,8,9])],
            ...     'third': 34
            ... })
            {
                'first': array([1, 2, 3]),
                'second': [array([4, 5, 6]), array([7, 8, 9])],
                'third': 34
            }

    See:
        :meth:`~poutyne.torch_apply` for supported types.
    """
    if copy:
        func = lambda t: t.detach().cpu().numpy().copy()
    else:
        func = lambda t: t.detach().cpu().numpy()
    return torch_apply(obj, func)


def torch_to(obj, *args, **kargs):
    return torch_apply(obj, lambda t: t.to(*args, **kargs))


def torch_apply(obj, func):
    """
    Apply a function to all tensors inside a Python object composed of the supported types.

    Supported types are: list, tuple and dict.

    Args:
        obj: The Python object to convert.
        func: The function to apply.

    Returns:
        A new Python object with the same structure as `obj` but where the tensors have been applied
        the function `func`. Not supported type are left as reference in the new object.
    """
    fn = lambda t: func(t) if torch.is_tensor(t) else t
    return _apply(obj, fn)


def _apply(obj, func):
    if isinstance(obj, (list, tuple)):
        return type(obj)(_apply(el, func) for el in obj)
    if isinstance(obj, dict):
        return {k: _apply(el, func) for k, el in obj.items()}
    return func(obj)


def _concat(obj):
    if isinstance(obj[0], (list, tuple)):
        return type(obj[0])(_concat(ele) for ele in zip(*obj))
    if isinstance(obj[0], dict):
        concat_dict = {}
        for key in obj[0].keys():
            concat_dict[key] = _concat([o[key] for o in obj])
        return concat_dict
    if isinstance(obj[0], np.ndarray) and obj[0].shape != ():
        return np.concatenate(obj)
    return obj


def numpy_to_torch(obj):
    """
    Convert to tensors all Numpy arrays inside a Python object composed of the supported types.

    Args:
        obj: The Python object to convert.

    Returns:
        A new Python object with the same structure as `obj` but where the Numpy arrays are now
        tensors. Not supported type are left as reference in the new object.

    Example:
        .. code-block:: python

            >>> from poutyne import numpy_to_torch
            >>> numpy_to_torch({
            ...     'first': np.array([1, 2, 3]),
            ...     'second':[np.array([4,5,6]), np.array([7,8,9])],
            ...     'third': 34
            ... })
            {
                'first': tensor([1, 2, 3]),
                'second': [tensor([4, 5, 6]), tensor([7, 8, 9])],
                'third': 34
            }


    """
    fn = lambda a: torch.from_numpy(a) if isinstance(a, np.ndarray) else a
    return _apply(obj, fn)


class TensorDataset(Dataset):
    """Dataset wrapping tensors.

    Each sample will be retrieved by indexing tensors along the first dimension.

    Arguments:
        *tensors (Tensor): tensors that have the same size of the first dimension.
    """

    def __init__(self, *tensors):
        super().__init__()
        self.tensors = tensors

        def _rabbit_hole(obj):
            if isinstance(obj, (list, tuple)):
                lengths = [_rabbit_hole(o) for o in obj]
                for length in lengths[1:]:
                    assert length == lengths[0]
                return lengths[0]
            return len(obj)

        self._len = _rabbit_hole(self.tensors)

    def __getitem__(self, index):
        def _rabbit_hole(obj, idx):
            if isinstance(obj, (list, tuple)):
                return type(obj)(_rabbit_hole(o, idx) for o in obj)
            return obj[idx]

        return _rabbit_hole(self.tensors, index)

    def __len__(self):
        return self._len


def set_seeds(seed):
    """
    Set Python, Numpy and Pytorch's random seeds in order to make
    the random number generation procedure deterministic and reproducible.

    Args:
        seed (int): The random number generation seed to use.
    """
    if seed is not None:
        random.seed(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)


def is_in_jupyter_notebook():
    # pylint: disable=import-outside-toplevel
    try:
        from IPython import get_ipython

        shell = get_ipython().__class__.__name__
        jupyter = shell in ['ZMQInteractiveShell', 'Shell']
    except ImportError:
        jupyter = False

    return jupyter



from unittest.mock import Mock, call

import torch
import torch.nn as nn

from poutyne import Model, LambdaCallback, Callback
from tests.framework.tools import some_data_tensor_generator
from tests.framework.model.base import ModelFittingTestCase


class LambdaTest(ModelFittingTestCase):
    def setUp(self):
        super().setUp()
        torch.manual_seed(42)
        self.pytorch_network = nn.Linear(1, 1)
        self.loss_function = nn.MSELoss()
        self.optimizer = torch.optim.Adam(self.pytorch_network.parameters(), lr=1e-3)
        self.model = Model(self.pytorch_network, self.optimizer, self.loss_function)

    def test_integration_zero_args(self):
        lambda_callback = LambdaCallback()

        train_generator = some_data_tensor_generator(LambdaTest.batch_size)
        valid_generator = some_data_tensor_generator(LambdaTest.batch_size)
        test_generator = some_data_tensor_generator(LambdaTest.batch_size)
        self.model.fit_generator(
            train_generator,
            valid_generator,
            epochs=LambdaTest.epochs,
            steps_per_epoch=LambdaTest.steps_per_epoch,
            validation_steps=LambdaTest.steps_per_epoch,
            callbacks=[lambda_callback],
        )

        num_steps = 10
        self.model.evaluate_generator(test_generator, steps=num_steps, callbacks=[lambda_callback])

    def test_with_only_on_epoch_end_arg(self):
        on_epoch_end = Mock()
        lambda_callback = LambdaCallback(on_epoch_end=on_epoch_end)

        train_generator = some_data_tensor_generator(LambdaTest.batch_size)
        valid_generator = some_data_tensor_generator(LambdaTest.batch_size)
        test_generator = some_data_tensor_generator(LambdaTest.batch_size)
        logs = self.model.fit_generator(
            train_generator,
            valid_generator,
            epochs=LambdaTest.epochs,
            steps_per_epoch=LambdaTest.steps_per_epoch,
            validation_steps=LambdaTest.steps_per_epoch,
            callbacks=[lambda_callback],
        )

        num_steps = 10
        self.model.evaluate_generator(test_generator, steps=num_steps, callbacks=[lambda_callback])

        expected_calls = [call(epoch_number, log) for epoch_number, log in enumerate(logs, 1)]
        actual_calls = on_epoch_end.mock_calls
        self.assertEqual(len(expected_calls), len(actual_calls))
        self.assertEqual(expected_calls, actual_calls)

    def test_lambda_test_calls(self):
        lambda_callback, mock_calls = self._get_lambda_callback_with_mock_args()
        num_steps = 10
        generator = some_data_tensor_generator(LambdaTest.batch_size)
        self.model.evaluate_generator(generator, steps=num_steps, callbacks=[lambda_callback, self.mock_callback])

        expected_calls = self.mock_callback.method_calls[2:]
        actual_calls = mock_calls.method_calls
        self.assertEqual(len(expected_calls), len(actual_calls))
        self.assertEqual(expected_calls, actual_calls)

    def test_lambda_train_calls(self):
        lambda_callback, mock_calls = self._get_lambda_callback_with_mock_args()
        train_generator = some_data_tensor_generator(LambdaTest.batch_size)
        valid_generator = some_data_tensor_generator(LambdaTest.batch_size)
        self.model.fit_generator(
            train_generator,
            valid_generator,
            epochs=LambdaTest.epochs,
            steps_per_epoch=LambdaTest.steps_per_epoch,
            validation_steps=LambdaTest.steps_per_epoch,
            callbacks=[lambda_callback, self.mock_callback],
        )

        expected_calls = self.mock_callback.method_calls[2:]
        actual_calls = mock_calls.method_calls
        self.assertEqual(len(expected_calls), len(actual_calls))
        self.assertEqual(expected_calls, actual_calls)

    def _get_lambda_callback_with_mock_args(self):
        mock_callback = Mock(spec=Callback())
        lambda_callback = LambdaCallback(
            on_epoch_begin=mock_callback.on_epoch_begin,
            on_epoch_end=mock_callback.on_epoch_end,
            on_train_batch_begin=mock_callback.on_train_batch_begin,
            on_train_batch_end=mock_callback.on_train_batch_end,
            on_valid_batch_begin=mock_callback.on_valid_batch_begin,
            on_valid_batch_end=mock_callback.on_valid_batch_end,
            on_test_batch_begin=mock_callback.on_test_batch_begin,
            on_test_batch_end=mock_callback.on_test_batch_end,
            on_train_begin=mock_callback.on_train_begin,
            on_train_end=mock_callback.on_train_end,
            on_valid_begin=mock_callback.on_valid_begin,
            on_valid_end=mock_callback.on_valid_end,
            on_test_begin=mock_callback.on_test_begin,
            on_test_end=mock_callback.on_test_end,
            on_backward_end=mock_callback.on_backward_end,
        )
        return lambda_callback, mock_callback
